c     # ------------------------------------------------------
c     # SRC_REACTION
c     #
c     # Definition of source term for convergence study.  We assume a solution
c     # of the form Qe(x,t) = A(t)Q(x).  The source term is then given by
c     #
c     #     S(x,t) = dQe/dt - \nabla^2 Qe = A'(t)Q(x) - A(t)\nabla^2 Q(x)
c     #
c     # Using this source term, we should get that the numerical solution
c     #    ||q(x,t) - Qe(x,t)|| = O(h^2)
c     #
c     # We pre-compute \nabla^2 Q(x) and Q(x) and store these in common blocks
c     # in 'para_exact.i'
c     #
c     # NOTE : Do not change the name 'src_reaction';  this is expected by
c     # the calling routine.
c     # ------------------------------------------------------

      subroutine src_diffusion(mx,my,meqn,t,q,lap)
      implicit none

      integer mx,my,meqn
      double precision t
      double precision lap(0:mx+1,0:my+1,*)
      double precision q(0:mx+1,0:my+1,*)

      integer mq, m, i,j, q_idx
      logical active_parabolic_term

      double precision beta(2)
      common /combeta/ beta

      do mq = 1,meqn
         if (active_parabolic_term(mq)) then
            m = q_idx(mq)
            do j = 1,my
               do i = 1,mx
                  lap(i,j,m) = beta(mq)*lap(i,j,m)
               enddo
            enddo
         endif
      enddo

      return
      end

      subroutine src_reaction(mx,my,meqn,t,q,rhs)
      implicit none

      integer mx,my,meqn
      double precision t
      double precision q(0:mx+1,0:my+1,*)
      double precision rhs(0:mx+1,0:my+1,*)

      integer mq, m, i,j, q_idx
      logical active_parabolic_term

      do mq = 1,meqn
         if (active_parabolic_term(mq)) then
            m = q_idx(mq)
            do j = 1,my
               do i = 1,mx
                  rhs(i,j,m) = 0
               enddo
            enddo
         endif
      enddo


      end

      subroutine src_reaction_irkc(i,j,nactive,t,q_pt,rhs_pt,
     &      compute_jac,jac)
      implicit none

      integer nactive,i,j
      double precision t
      double precision rhs_pt(nactive), q_pt(nactive)
      double precision jac(nactive,nactive)
      logical compute_jac
      integer m, ii,jj

      integer mx,my,meqn

      do ii = 1,nactive
         rhs_pt(ii) = 0
         if (compute_jac) then
            do jj = 1,nactive
               jac(ii,jj) = 0
            enddo
         endif
      enddo

      end


c     # -------------------------------------------------------
c     # APPLY_LB
c     #
c     # Apply the diffusion operator. Boundary data has aleady
c     # been supplied to qp and qd.
c     # -------------------------------------------------------
      subroutine apply_lb(mx,my,mbc,meqn,coeffs,area,qp,qd,Ap)
      implicit none

      integer mx,my, meqn,mbc

      double precision qp(0:mx+1,0:my+1,*)
      double precision qd(mx+1,my+1,*)
      double precision ap(0:mx+1,0:my+1,*)

      double precision coeffs(0:mx+1,0:my+1,2,2)
      double precision area(-mbc:mx+mbc+1,-mbc:my+mbc+1)

      integer i,j,m, mq, q_idx
      double precision fimh, gjmh
      logical active_parabolic_term


c     # Loop over each equation
      do mq = 1,meqn
         if (active_parabolic_term(mq)) then
            m = q_idx(mq)
c           # Compute fluxes at vertical edges of all mesh cells
c           # Add contributions to L(i,j) (L == Ap)
            do j = 1,my
               do i = 1,mx+1
                  fimh = coeffs(i,j,1,1)*(qp(i,j,m)   - qp(i-1,j,m)) +
     &                  coeffs(i,j,1,2)*(qd(i,j+1,m) - qd(i,j,m))
                  Ap(i,j,m)   = -fimh
                  Ap(i-1,j,m) = Ap(i-1,j,m) + fimh
               enddo
            enddo

c           # Compute fluxes at bottom edges of each mesh cell; add
c           contributions
c           # to two cells adjacent to this edge.
            do j = 1,my+1
               do i = 1,mx
                  gjmh = coeffs(i,j,2,1)*(qp(i,j,m)   - qp(i,j-1,m)) +
     &                  coeffs(i,j,2,2)*(qd(i+1,j,m) - qd(i,j,m))
                  Ap(i,j,m)   = Ap(i,j,m) - gjmh
                  Ap(i,j-1,m) = Ap(i,j-1,m) + gjmh
               enddo
            enddo

c           # Divide through by area to get an approximation
c           # to the pointwise Laplacian
            do j = 1,my
               do i = 1,mx
                  Ap(i,j,m) = Ap(i,j,m)/area(i,j)
               enddo
            enddo
         endif
      enddo

      call increment_laplacian_call_counter()

      end

c     # -------------------------------------------------------
c     # COMPUTE_DUAL_SPHERE
c     #
c     # Apply the averaging operator to get dual values
c     #
c     # For smooth mappings dual values can be computed as the
c     # arithmetic average of primal nodes.
c     #
c     # The averaging below was designed for the disk mapping
c     # and so we we have to treat the diagonals specially.
c     # -------------------------------------------------------
      subroutine compute_dual_sphere(mx,my,meqn,qp,qd)
      implicit none

      integer mx,my,meqn

      double precision qp(0:mx+1,0:my+1,*)
      double precision qd(mx+1,my+1,*)

      integer order
      integer i,j,m, q_idx, mq
      logical active_parabolic_term

c     # To get second order, average dual cells along the diagonal
c     # using only primal values on the diagonal.
      do mq = 1,meqn
         if (active_parabolic_term(mq)) then
            m = q_idx(mq)
            do j = 2,my
               do i = 2,mx
                  if (i .eq. j .or. i - mx/2 .eq. j) then
                     qd(i,j,m) = 0.5d0*(qp(i,j,m) + qp(i-1,j-1,m))
                  elseif (abs(i+j) .eq. my+2 .or.
     &                     abs(i-mx/2+j) .eq. my+2) then
                     qd(i,j,m) = 0.5d0*(qp(i-1,j,m) + qp(i,j-1,m))
                  else
                     qd(i,j,m) = 0.25d0*(qp(i-1,j,m) +
     &                     qp(i-1,j-1,m) + qp(i,j-1,m) + qp(i,j,m))
                  endif
               enddo
            enddo
         endif
      enddo


      end
