!!
!! PHYS2COMP
!!
!! Converts physical (xp,yp,zp) coordinates to (xc,yc) coordinates used in the two-patch sphere grid, i.e.
!! (xi,eta) coordinates.
!!
!!                  CALL phys2comp(xp,yp,zp,xc,yc)
!!
!! We do not invert our mapping directly, but rather model the mapping locally using a bilinear
!! approximation.  We then find the (xc,yc) that minimizes the distance
!!
!!                  F(xc,yc) = 0.5*||Ta(xc,yc) - Tp||^2
!!
!! where Tp = (xp,yp,zp), and Ta(xc,yc) approximate bilinear mapping.  The 'cellave' routine is used to
!! find a box (ll,ur) which contains (in most cases)  the (xc,yc) coordinate.   The corners of this box
!! in physical space are then used to construct the bilinear approximation.
!!
!! Some care has to be taken to avoid the non-smooth regions of the mapping.
!!
!! This routine depends on 'cellave.f' and the sphere mapping 'mapc2m.f'.
!!

SUBROUTINE phys2comp(xp,yp,zp,xc,yc,tol)
  IMPLICIT NONE

  DOUBLE PRECISION :: xp,yp,zp,xc,yc, tol

  DOUBLE PRECISION :: ll(2), ur(2), u, v
  DOUBLE PRECISION :: compute_err_p2c, err, s
  LOGICAL :: debug
  LOGICAL :: is_diag_box

  COMMON /com_debug_p2c/ debug

!! Get box containing (xc,yc)
  CALL get_box_p2c(xp,yp,zp,ll,ur,tol)

!! Set common blocks needed for bilinear approximation Ta(xc,yc) in box (ll,ur)
  CALL setup_map_p2c(ll,ur)

!! Find (xc,yc) which minimizes distance between Tp=(xp,yp,zp) and Ta(xc,yc)
  CALL solve_newton_p2c(u,v)

  xc = ll(1) + (ur(1)-ll(1))*u
  yc = ll(2) + (ur(2)-ll(2))*v

  IF (debug) THEN
     IF (is_diag_box(ll,ur)) THEN
        s = 1
     ELSE
        s = -1
     ENDIF

     err =  compute_err_p2c(xc,yc)

     OPEN(10,file='xy.out')
     WRITE(10,'(2F16.8)') ll(1), ll(2)
     WRITE(10,'(2F16.8)') ur(1), ur(2)
     WRITE(10,'(2F16.8)') xc, yc
     WRITE(10,'(3E16.8)') err, s
     CLOSE(10)
  ENDIF

END SUBROUTINE phys2comp

SUBROUTINE get_box_p2c(xp,yp,zp,ll_out,ur_out,tol)
  IMPLICIT NONE

  DOUBLE PRECISION :: ll_out(2), ur_out(2), tol
  DOUBLE PRECISION :: xp, yp, zp

  INTEGER :: i,j, m

  INTEGER :: rf, i1, j1
  DOUBLE PRECISION :: ll(2), ur(2), h, u, v, err, compute_err_p2c
  DOUBLE PRECISION :: xc, yc, xlow, ylow
  LOGICAL :: contains_coord

  DOUBLE PRECISION :: xp_com, yp_com, zp_com
  LOGICAL :: debug
  LOGICAL :: contains_p, contains_point
  LOGICAL :: is_diag_box, okay

!! These common blocks are used to communicate with function fdisc, below
  COMMON /com_physpoint_p2c/ xp_com, yp_com, zp_com
  COMMON /com_debug/ debug

  xp_com = xp
  yp_com = yp
  zp_com = zp

!! Coarse grid approximation to locate x
  IF (zp .LT. 0) THEN
     ll(1) = -3.d0
     ll(2) = -1.d0
  ELSE
     ll(1) = -1.d0
     ll(2) = -1.d0
  ENDIF
  rf = 4
  h = 2.d0/rf

  err = 1.d0
  DO WHILE (err .GT. tol .AND. rf .LT. 2**12)
     rf = 2*rf
     h = h/2.d0
     xlow = ll(1)
     ylow = ll(2)
     outer : DO i = 1,rf
        inner : DO j = 1,rf
           ll(1) = xlow + (i-1)*h
           ll(2) = ylow + (j-1)*h
           ur(1) = ll(1) + h
           ur(2) = ll(2) + h

           contains_p = contains_point(ll,ur)
           IF (contains_p) THEN
              EXIT outer
           ENDIF
        ENDDO inner
     ENDDO outer
     !! Set common blocks needed for bilinear approximation Ta(xc,yc) in box (ll,ur)
     CALL setup_map_p2c(ll,ur)

     !! Find (xc,yc) which minimizes distance between Tp=(xp,yp,zp) and Ta(xc,yc)
     CALL solve_newton_p2c(u,v)
     xc = ll(1) + u*h
     yc = ll(2) + v*h
     err = compute_err_p2c(xc,yc)
  ENDDO

  DO m = 1,2
     ll_out(m) = ll(m)
     ur_out(m) = ur(m)
  ENDDO


END SUBROUTINE get_box_p2c

DOUBLE PRECISION FUNCTION compute_err_p2c(xc,yc)
  IMPLICIT NONE

  DOUBLE PRECISION xc,yc,xp,yp,zp, xp_com, yp_com, zp_com

  COMMON /com_physpoint_p2c/ xp_com, yp_com, zp_com

  CALL mapc2m(xc,yc,xp,yp,zp)

  compute_err_p2c = SQRT((xp - xp_com)**2 + (yp - yp_com)**2 + (zp - zp_com)**2)

END FUNCTION compute_err_p2c


LOGICAL FUNCTION contains_point(ll,ur)
  IMPLICIT NONE

  DOUBLE PRECISION :: ll(2), ur(2)

  LOGICAL contains_x, contains_y, contains_z
  LOGICAL contains_coord

  contains_x = contains_coord(ll,ur,1)
  contains_y = contains_coord(ll,ur,2)
  contains_z = contains_coord(ll,ur,3)

  contains_point = contains_x .and. contains_y .and. contains_z

END FUNCTION contains_point

LOGICAL FUNCTION contains_coord(ll,ur,ic)
  IMPLICIT NONE

  DOUBLE PRECISION :: ll(2), ur(2)
  INTEGER :: ic

  DOUBLE PRECISION :: dx, dy, w
  INTEGER :: ic_com

  COMMON /com_ic_p2c/ ic_com

  dx = ur(1) - ll(1)
  dy = ur(2) - ll(2)

  ic_com = ic
  CALL cellave(ll(1),ll(2),dx,dy,w)

  contains_coord = 0 .LT. w .AND. w .LT. 1.d0

END FUNCTION contains_coord


!! This is called by 'cellave', and so has to have this name.
DOUBLE PRECISION FUNCTION fdisc(xc,yc)
  IMPLICIT NONE

  DOUBLE PRECISION :: xc,yc, xp_est,yp_est,zp_est
  DOUBLE PRECISION :: xp_com, yp_com, zp_com
  INTEGER :: ic, ic_com

  COMMON /com_physpoint_p2c/ xp_com, yp_com, zp_com
  COMMON /com_ic_p2c/ ic_com

  ic = ic_com

  CALL mapc2m(xc,yc,xp_est,yp_est,zp_est)

  IF (ic .EQ. 1) THEN
     fdisc = xp_est - xp_com
  ELSEIF (ic .eq. 2) then
     fdisc = yp_est - yp_com
  ELSEIF (ic .EQ. 3) THEN
     fdisc = zp_est - zp_com
  ENDIF

END FUNCTION fdisc

LOGICAL FUNCTION is_diag_box(ll,ur)
  IMPLICIT NONE

  DOUBLE PRECISION :: ll(2), ur(2), lr(2), ul(2)

  DOUBLE PRECISION :: s, b, h

  is_diag_box = .FALSE.

  h = ur(1) - ll(1)
  lr(1) = ll(1) + h
  lr(2) = ll(2)

  ul(1) = ur(1) - h
  ul(2) = ur(2)

  s = (ur(2)-ll(2))/(ur(1)-ll(1))
  b = -s*ll(1) + ll(2)
  IF (ABS(b - 2.d0) .LT. 1d-10 .OR. ABS(b) .LT. 1e-10) THEN
     is_diag_box = .TRUE.
     RETURN
  ENDIF

  s = (ul(2)-lr(2))/(ul(1)-lr(1))
  b = -s*lr(1) + lr(2)

  IF (ABS(b + 2.d0) .LT. 1d-10 .OR. ABS(b) .LT. 1e-10) THEN
     is_diag_box = .TRUE.
  ENDIF



END FUNCTION is_diag_box


!! Set-up bilinear approximation to sphere mapping
SUBROUTINE setup_map_p2c(ll,ur)
  IMPLICIT NONE

  DOUBLE PRECISION :: ll(2), ur(2)

  INTEGER :: ii, jj, m
  DOUBLE PRECISION :: quad(0:1,0:1,3), xp, yp, zp, xc, yc, xlow, ylow, dx, dy
  DOUBLE PRECISION :: a00(3), a01(3), a10(3), a11(3)

  COMMON /com_map_p2c/ a00, a01, a10, a11

  xlow = ll(1)
  ylow = ll(2)
  dx = ur(1) - ll(1)
  dy = ur(2) - ll(2)

  DO ii = 0,1
     DO jj = 0,1
        xc = xlow + ii*dx
        yc = ylow + jj*dy
        CALL mapc2m(xc,yc,xp,yp,zp)
        quad(ii,jj,1) = xp
        quad(ii,jj,2) = yp
        quad(ii,jj,3) = zp
     ENDDO
  ENDDO

  DO m = 1,3
     a00(m) = quad(0,0,m)
     a01(m) = quad(1,0,m) - quad(0,0,m)
     a10(m) = quad(0,1,m) - quad(0,0,m)
     a11(m) = quad(1,1,m) - quad(1,0,m) - (quad(0,1,m) - quad(0,0,m))
  ENDDO

END SUBROUTINE setup_map_p2c

!!
!! Nonlinear system grad F = 0 that is solved to minimize F(xc,yc).  Note that we assume that
!! xc = ll(1) + u*(ur(1)-ll(1)) and yc = ll(2) + v*(ur(2)-ll(2))
!!
SUBROUTINE f_nonlin_p2c(u,v,gradf)
  IMPLICIT NONE

  DOUBLE PRECISION :: u,v, gradf(2)

  DOUBLE PRECISION :: t(3), tp(3)
  DOUBLE PRECISION :: a00(3), a01(3), a10(3), a11(3), txi(3), teta(3)
  DOUBLE PRECISION :: xp_com, yp_com, zp_com
  INTEGER :: ii, jj, m

  COMMON /com_map_p2c/ a00, a01, a10, a11
  COMMON /com_physpoint_p2c/ xp_com, yp_com, zp_com

  tp(1) = xp_com
  tp(2) = yp_com
  tp(3) = zp_com

  DO m = 1,3
     t(m) = a00(m) + a01(m)*u + a10(m)*v + a11(m)*u*v
     txi(m) = a01(m) + a11(m)*v
     teta(m) = a10(m) + a11(m)*u
  ENDDO

  gradf(1) = (t(1)-tp(1))*txi(1)  + (t(2)-tp(2))*txi(2)  + (t(3)-tp(3))*txi(3)
  gradf(2) = (t(1)-tp(1))*teta(1) + (t(2)-tp(2))*teta(2) + (t(3)-tp(3))*teta(3)

END SUBROUTINE f_nonlin_p2c

!! Inverse of the Hessian matrix of F(xc,yc) = 0.5||Ta(xc,yc) - Tp||^2
SUBROUTINE hessian_inv_p2c(u,v,hess_inv)
  IMPLICIT NONE

  DOUBLE PRECISION :: u,v,hess(2,2)
  DOUBLE PRECISION :: t(3), tp(3), txi(3), teta(3)
  DOUBLE PRECISION :: a00(3), a01(3), a10(3), a11(3)
  DOUBLE PRECISION :: hess_det, hess_inv(2,2)
  DOUBLE PRECISION :: xp_com, yp_com, zp_com
  INTEGER :: m

  COMMON /com_map_p2c/ a00, a01, a10, a11
  COMMON /com_physpoint_p2c/ xp_com, yp_com, zp_com

  tp(1) = xp_com
  tp(2) = yp_com
  tp(3) = zp_com

  DO m = 1,3
     t(m) = a00(m) + a01(m)*u + a10(m)*v + a11(m)*u*v
     txi(m) = a01(m) + a11(m)*v
     teta(m) = a10(m) + a11(m)*u
  ENDDO

  hess(1,1) = txi(1)**2 + txi(2)**2 + txi(3)**2
  hess(2,2) = teta(1)**2 + teta(2)**2 + teta(3)**2
  hess(1,2) = (t(1) - tp(1))*a11(1) + (t(2)-tp(2))*a11(2) + (t(3)-tp(3))*a11(3) &
       & + txi(1)*teta(1) + txi(2)*teta(2) + txi(3)*teta(3)
  hess(2,1) = hess(1,2)

  hess_det = hess(1,1)*hess(2,2) - hess(1,2)*hess(2,1)
  hess_inv(1,1) = hess(2,2)/hess_det
  hess_inv(1,2) = -hess(1,2)/hess_det
  hess_inv(2,1) = -hess(2,1)/hess_det
  hess_inv(2,2) = hess(1,1)/hess_det


END SUBROUTINE hessian_inv_p2c

SUBROUTINE solve_newton_p2c(u_est, v_est)
  IMPLICIT NONE

  DOUBLE PRECISION :: u_est, v_est
  LOGICAL :: debug

  INTEGER :: k, kmax
  DOUBLE PRECISION :: uk, vk, ukp1, vkp1, gradf(2)
  DOUBLE PRECISION :: tol, gradn, hess_inv(2,2)

  COMMON /com_debug_p2c/ debug

  IF (debug) THEN
     CALL write_line_p2c(6,55)
     WRITE(6,50) 'uk        vk       f(u,v)        '
     CALL write_line_p2c(6,55)
50   FORMAT(A)
  ENDIF

  kmax = 100
  tol = 1d-10

  uk = 0.5d0
  vk = 0.5d0

  DO k = 1,kmax
     CALL f_nonlin_p2c(uk,vk,gradf)
     gradn = sqrt(gradf(1)*gradf(1) + gradf(2)*gradf(2))

     IF (debug) THEN
        WRITE(6,100) k, uk,vk,gradn
100     FORMAT(I5,2F16.8, E16.8)
     ENDIF

     IF (gradn .LT. tol) THEN
        IF (debug) THEN
           CALL write_line_p2c(6,55)
           WRITE(6,100) k, uk, vk, gradn
        ENDIF
        u_est = uk
        v_est = vk
        RETURN
     ENDIF

     CALL hessian_inv_p2c(uk,vk,hess_inv)

     ukp1 = uk - hess_inv(1,1)*gradf(1) - hess_inv(1,2)*gradf(2)
     vkp1 = vk - hess_inv(2,1)*gradf(1) - hess_inv(2,2)*gradf(2)

     uk = ukp1
     vk = vkp1
  ENDDO

  WRITE(6,*) 'Maximum iterations exceeded'
  STOP

END SUBROUTINE solve_newton_p2c


SUBROUTINE write_line_p2c(iunit,len)
  IMPLICIT NONE

  INTEGER iunit,len,i
  CHARACTER(1000) line

  IF (len .GT. 1000) THEN
     WRITE(6,*) 'write_line : len > 1000'
     STOP
  ENDIF

  DO i = 1,len
     line(i:i) = '-'
  ENDDO

  WRITE(iunit,'(A)') line(1:len)

END SUBROUTINE write_line_p2c
