c     # -------------------------------------------------------
c     # Setup system for solving for dual values at the equator.
c     #
c     # This requires that the tridiagonal system has been
c     # set up by 'initialize_bc(...)'
c     #
c     # ----------------------------------------------------------

      subroutine apply_equator(mx,my,meqn,work_tri,mwork_tri,
     &      ipiv,qp,qd)
      implicit none

      integer mx,my,meqn,mwork_tri
c     # The last dimension is the number of equations
c     # containing parabolic terms; this is <= meqn
      double precision qp(0:mx+1,0:my+1,*)
      double precision qd(0:mx+1,0:my+1,*)

      double precision work_tri(mwork_tri)
      integer ipiv(2*(mx/2+my))

      integer m,mq
      logical active_parabolic_term
      integer q_idx

      do mq = 1,meqn
         if (active_parabolic_term(mq)) then
            m = q_idx(mq)
            call apply_equator_internal(mx,my,work_tri,mwork_tri,
     &            ipiv,qp(0,0,m),qd(0,0,m))
         endif
      enddo

      end


      subroutine apply_equator_internal(mx,my,work_tri,mwork_tri,ipiv,
     &      qp,qd)
      implicit none

      integer mx,my,mwork_tri
      double precision qp(0:mx+1,0:my+1)
      double precision qd(mx+1,my+1)
      double precision work_tri(mwork_tri)
      double precision ipiv(2*(mx/2+my))

      double precision fa

      integer pcval
      integer trisys_du, trisys_d, trisys_dl, trisys_du2, qh
      integer coeffs_alpha, dval, pval, rhs, rv
      integer xpi,ypi,zpi,xdi,ydi,zdi,tv

      integer ntri, nrhs, ldb, info
      character trans

c     # ----------------------------------------------------
c     # Get pointers needed for tridiagonal system and other
c     # set in initialization step
c     # ----------------------------------------------------
      call get_tri_ptrs(mwork_tri,mx,my,pcval,
     &      trisys_du,trisys_d,trisys_dl,trisys_du2,qh,
     &      coeffs_alpha,xpi,ypi,zpi,xdi,ydi,zdi,tv,
     &      pval,dval,rv,rhs)

c     # ----------------------------------------------------
c     # Get time dependent right hand side and scalar fa
c     # ----------------------------------------------------
      call get_tri_rhs(mx,my,work_tri(pcval),work_tri(pval),
     &      qp,work_tri(rv),work_tri(rhs),fa)

c     # ----------------------------------------------------
c     # Solve tridiagonal system to get dual values at the
c     # equator
c     # ----------------------------------------------------
      trans = 'n'
      nrhs = 1
      ntri = 2*(mx/2 + my) - 1
      ldb = ntri
      call dgttrs(trans, ntri,nrhs,work_tri(trisys_dl),
     &      work_tri(trisys_d),work_tri(trisys_du),
     &      work_tri(trisys_du2),ipiv,
     &      work_tri(rhs),ldb,info)

c     # ----------------------------------------------------
c     # Assign dual values at the equator. For the
c     # sphere, we don't use the primal values on the
c     # equator, since they are true ghost values, and are
c     # set using matching conditions with other cells in the
c     # interior of the domain.
c     # ----------------------------------------------------
      call assign_equator_dual(mx,my,work_tri(qh),work_tri(rhs),
     &      work_tri(coeffs_alpha),fa,work_tri(dval),qd)

      end

      subroutine get_tri_rhs(mx,my,pcval,pval,qp,rv,rhs,fa)
      implicit none

      integer mx,my

      double precision pcval(2*(mx/2+my),2)
      double precision pval(2*(mx/2+my),2)
      double precision qp(0:mx+1,0:my+1)

      double precision rv(0:2*(mx/2 + my))
      double precision rhs(2*(mx/2+my))
      double precision fa

      integer i, j, k, np
      double precision pl, pu

      call match_p(mx,my,qp)

      np = 2*(mx/2 + my)

c     # Index 1 : top
c     # Index 2 : bottom
      k = 1
      do i = mx/2+1,mx
         pval(k,1) = qp(i,1)
         pval(k,2) = qp(i,0)
         k = k + 1
      enddo

      do j = 1,my
         pval(k,1) = qp(mx,j)
         pval(k,2) = qp(mx+1,j)
         k = k + 1
      enddo

      do i = mx,mx/2+1,-1
         pval(k,1) = qp(i,my)
         pval(k,2) = qp(i,my+1)
         k = k + 1
      enddo

      do j = my,1,-1
         pval(k,1) = qp(mx/2+1,j)
         pval(k,2) = qp(mx/2,j)
         k = k + 1
      enddo

      do k = 1,np
         pu = pcval(k,1)
         pl = pcval(k,2)
         rv(k) = (pu*pval(k,1) + pl*pval(k,2))/(pu + pl)
      enddo
      rv(0) = rv(np)

      do k = 1,np
         rhs(k) = -(rv(k-1) + rv(k))
      enddo
      fa = rhs(np)

      end

      subroutine assign_equator_dual(mx,my,qh,qinh,coeffs_alpha,fa,
     &      dval,qd)
      implicit none

      integer mx,my
      double precision   qh(2*(mx/2+my))
      double precision qinh(2*(mx/2+my))
      double precision dval(2*(mx/2+my))

      double precision qd(mx+1,my+1)

      double precision fa, alpha, coeffs_alpha(3)

      integer i,j,k,np

      np = 2*(mx/2 + my)

      alpha = coeffs_alpha(1)*fa + coeffs_alpha(2)*qinh(1) +
     &      coeffs_alpha(3)*qinh(np-1)

      if (1./alpha .eq. 0) then
         alpha = 0
      endif

      do k = 1,np-1
         dval(k) = qinh(k) + alpha*qh(k)
      enddo
      dval(np) = alpha

c     # Set dual boundary values
      k = 1
      do i = mx/2+1,mx
         qd(i,1) = dval(k)
         k = k + 1
      enddo

      do j = 1,my
         qd(mx+1,j) = dval(k)
         k = k + 1
      enddo

      do i = mx+1,mx/2+2,-1
         qd(i,my+1) = dval(k)
         k = k + 1
      enddo

      do j = my+1,2,-1
         qd(mx/2+1,j) = dval(k)
         k = k + 1
      enddo

      call match_d(mx,my,qd)
  100 format(I5,3F24.16)
      end


c     # ----------------------------------------------------------------
c     # Setup the tridiagonal system
c     # ----------------------------------------------------------------
      subroutine initialize_equator(mx,my,xlower,ylower,dx,dy,
     &      xp,yp,zp,xd,yd,zd,work_tri,mwork_tri,ipiv)
      implicit none

      integer mx,my,mwork_tri
      double precision xlower,ylower,dx,dy

      double precision xp(0:mx+1,0:my+1)
      double precision yp(0:mx+1,0:my+1)
      double precision zp(0:mx+1,0:my+1)

      double precision xd(mx+1,my+1)
      double precision yd(mx+1,my+1)
      double precision zd(mx+1,my+1)

      double precision work_tri(mwork_tri)
      integer ipiv(2*(mx/2 + my))

      integer trisys_du,trisys_d,trisys_dl,trisys_du2
      integer pcval, qh, coeffs_alpha
      integer xpi, ypi, zpi, xdi,ydi,zdi
      integer pval,dval,rv,rhs
      integer tv

c     # ----------------------------------------------------
c     # Get pointers needed for tridiagonal system and other
c     # set in initialization step
c     # ----------------------------------------------------
      call get_tri_ptrs(mwork_tri,mx,my,pcval,
     &      trisys_du,trisys_d,trisys_dl,trisys_du2,qh,
     &      coeffs_alpha,xpi,ypi,zpi,xdi,ydi,zdi,tv,
     &      pval,dval,rv,rhs)

c     # ----------------------------------------------------
c     # Get coefficients for tridiagonal system
c     # ----------------------------------------------------
      call setup_equator_mesh(mx,my,
     &      xlower,ylower,dx,dy,xp,yp,zp,xd,yd,zd,
     &      work_tri(xpi),work_tri(ypi),work_tri(zpi),
     &      work_tri(xdi),work_tri(ydi),work_tri(zdi))

      call compute_trisys_coeffs(mx,my,
     &      work_tri(xpi),work_tri(ypi),work_tri(zpi),
     &      work_tri(xdi),work_tri(ydi),work_tri(zdi),
     &      work_tri(pcval),work_tri(tv))

c     # ----------------------------------------------------
c     # Get coefficients for tridiagonal system
c     # This could also be used for the bc_phys.f routine.
c     # ----------------------------------------------------
      call construct_sys(mx,my,work_tri(tv),
     &      work_tri(trisys_dl),work_tri(trisys_d),
     &      work_tri(trisys_du),work_tri(trisys_du2),
     &      ipiv,work_tri(qh),
     &      work_tri(coeffs_alpha))

      end

      subroutine construct_sys(mx,my,tv,
     &      trisys_dl,trisys_d,trisys_du,trisys_du2,
     &      ipiv, qh, coeffs_alpha)
      implicit none

      integer mx,my
      double precision  trisys_dl(2*(mx/2+my))
      double precision   trisys_d(2*(mx/2+my))
      double precision  trisys_du(2*(mx/2+my))
      double precision trisys_du2(2*(mx/2+my))
      integer                ipiv(2*(mx/2+my))

      double precision     qh(2*(mx/2+my))
      double precision coeffs_alpha(3)

      double precision tv(0:2*(mx/2 + my))

      double precision v(2),w(2), u
      integer k, km, np, ntri, ldb, nrhs, info
      character trans

      np = 2*(mx/2+my)

      do k = 1,np-1
         trisys_dl(k) = tv(k)
      enddo

      do k = 1,np
         trisys_d(k) = tv(k) - tv(k-1) - 2.d0
      enddo

      do k = 1,np-1
         trisys_du(k) = -tv(k)
      enddo

c     # Define vectors w and v. We only store v(1) and v(np-1); the rest
c     # are zero.  Same for w.
      w(1) = tv(np)
      w(2) = -tv(np-1)
      v(1) = -tv(np)
      v(2) = tv(np-1)
      u = tv(np) - tv(np-1) - 2.d0

c     # Factor tridiagonal system.
      ntri = np - 1
      call dgttrf(ntri,trisys_dl, trisys_d, trisys_du,
     &      trisys_du2, ipiv,info)

      if (info .ne. 0) then
         write(6,'(A,I5)') 'construct_sys(1) : info .ne. 0; ',
     &         info
         stop
      endif

c     # w - equation (2.41)
      do k = 1,np-1
         qh(k) = 0.d0
      enddo
      qh(1)    = -w(1)
      qh(np-1) = -w(2)

  100 format(I5,4E24.16)

      trans = 'n'
      ldb      = ntri
      nrhs     = 1
      call dgttrs(trans,ntri,nrhs,trisys_dl,trisys_d,
     &      trisys_du,trisys_du2,ipiv,qh,ldb,info)

      if (info .ne. 0) then
         write(6,'(A,I5)') 'construct_sys(2) : info .ne. 0; ',
     &         info
         stop
      endif

      coeffs_alpha(1) = 1.d0/(u + v(1)*qh(1) + v(2)*qh(np-1))
      coeffs_alpha(2) = -coeffs_alpha(1)*v(1)
      coeffs_alpha(3) = -coeffs_alpha(1)*v(2)

      end

c     # -------------------------------------------------------
c     # Setup grid at the equator needed to compute coefficients.
c     # -------------------------------------------------------
      subroutine setup_equator_mesh(mx,my,xlower,ylower,dx,dy,
     &      xp,yp,zp,xd,yd,zd,xpi,ypi,zpi,xdi,ydi,zdi)
      implicit none

      integer mx,my
      double precision xlower,ylower,dx,dy

      double precision xp(0:mx+1,0:my+1)
      double precision yp(0:mx+1,0:my+1)
      double precision zp(0:mx+1,0:my+1)

      double precision xd(mx+1,my+1)
      double precision yd(mx+1,my+1)
      double precision zd(mx+1,my+1)

      double precision xpi(2*(mx/2 + my),-1:1)
      double precision ypi(2*(mx/2 + my),-1:1)
      double precision zpi(2*(mx/2 + my),-1:1)

      double precision xdi(2*(mx/2 + my))
      double precision ydi(2*(mx/2 + my))
      double precision zdi(2*(mx/2 + my))

      integer i,j,n
      double precision xc, yc, xd1, yd1, zd1
      double precision xp1, yp1,zp1
      double precision xupper, yupper

      double precision d1, d2, d3
      integer np1, np

      if (mx/2 .ne. my) then
         write(6,*) 'setup_mesh_interal : mx/2 .ne. my'
         stop
      endif

      call match_p(mx,my,xp)
      call match_p(mx,my,yp)
      call match_p(mx,my,zp)

      xupper = xlower + mx*dx
      yupper = ylower + my*dy

      n = 1
      do i = mx/2 + 1, mx
         do j = -1,1,2
            xpi(n,j) = xp(i,(j+1)/2)
            ypi(n,j) = yp(i,(j+1)/2)
            zpi(n,j) = zp(i,(j+1)/2)
         enddo
         xc = xlower + (i-0.5)*dx
         yc = ylower
         call mapc2m(xc,yc,xpi(n,0), ypi(n,0),zpi(n,0))

         n = n + 1
      enddo

      do j = 1,my
         do i = -1,1,2
            xpi(n,i) = xp((1-i)/2 + mx,j)
            ypi(n,i) = yp((1-i)/2 + mx,j)
            zpi(n,i) = zp((1-i)/2 + mx,j)
         enddo
         xc = xupper
         yc = ylower + (j-0.5)*dy
         call mapc2m(xc,yc,xpi(n,0), ypi(n,0),zpi(n,0))
         n = n + 1
      enddo

      do i = mx,mx/2+1,-1
         do j = -1,1,2
            xpi(n,j) = xp(i,(1-j)/2 + my)
            ypi(n,j) = yp(i,(1-j)/2 + my)
            zpi(n,j) = zp(i,(1-j)/2 + my)
         enddo
         xc = xlower + (i-0.5)*dx
         yc = yupper
         call mapc2m(xc,yc,xpi(n,0), ypi(n,0),zpi(n,0))
         n = n + 1
      enddo

      do j = my,1,-1
         do i = -1,1,2
            xpi(n,i) = xp((i+1+mx)/2,j)
            ypi(n,i) = yp((i+1+mx)/2,j)
            zpi(n,i) = zp((i+1+mx)/2,j)
         enddo
         xc = xlower + mx*dx/2.d0
         yc = ylower + (j-0.5)*dy
         call mapc2m(xc,yc,xpi(n,0), ypi(n,0),zpi(n,0))

         n = n + 1
      enddo

c     # Dual cells
      n = 1
      do i = mx/2+1,mx
         xdi(n) = xd(i,1)
         ydi(n) = yd(i,1)
         zdi(n) = zd(i,1)
         n = n + 1
      enddo

      do j = 1,my
         xdi(n) = xd(mx+1,j)
         ydi(n) = yd(mx+1,j)
         zdi(n) = zd(mx+1,j)
         n = n + 1
      enddo

      do i = mx+1,mx/2+2,-1
         xdi(n) = xd(i,my+1)
         ydi(n) = yd(i,my+1)
         zdi(n) = zd(i,my+1)
         n = n + 1
      enddo

      do j = my+1,2,-1
         xdi(n) = xd(mx/2+1,j)
         ydi(n) = yd(mx/2+1,j)
         zdi(n) = zd(mx/2+1,j)
         n = n + 1
      enddo

  100 format(I5,3E16.8)

      end

c     # ----------------------------------------------------------------
c     # Setup the coefficients needed for the boundary conditions operator
c     # ----------------------------------------------------------------
      subroutine compute_trisys_coeffs(mx,my,xpi,ypi,zpi,
     &      xdi,ydi,zdi,pcval,tv)
      implicit none

      integer mx,my
      double precision xpi(2*(mx/2 + my),-1:1)
      double precision ypi(2*(mx/2 + my),-1:1)
      double precision zpi(2*(mx/2 + my),-1:1)

      double precision xdi(2*(mx/2 + my))
      double precision ydi(2*(mx/2 + my))
      double precision zdi(2*(mx/2 + my))

      double precision pcval(2*(mx/2 + my),2)

      double precision tv(0:2*(mx/2 + my))

      integer i,k,m, np, kp1
      double precision taup_head(3),taup_tail(3), taup(3), sp
      double precision taud_head(3),taud_tail(3), taud(3), sd
      double precision d_points(2,3)
      double precision p_points(3)
      double precision a1, a2, ca, sa, cta

      double precision uu, ul, pu, pl, u(2), p(2)

      np = 2*(mx/2 + my)

c     % Loop over each primary cell
      do k = 1,np
         do i = -1,1,2
            taud_tail(1) = xpi(k,i)
            taud_tail(2) = ypi(k,i)
            taud_tail(3) = zpi(k,i)

c           # These are the points on nodes of the primary cell
            if (k .eq. np) then
               kp1 = 1
            else
               kp1 = k + 1
            endif
            d_points(1,1) = xdi(k)
            d_points(2,1) = xdi(kp1)

            d_points(1,2) = ydi(k)
            d_points(2,2) = ydi(kp1)

            d_points(1,3) = zdi(k)
            d_points(2,3) = zdi(kp1)

c           # These are the points at the end of each arm
            p_points(1) = xpi(k,0)
            p_points(2) = ypi(k,0)
            p_points(3) = zpi(k,0)

            sp = 0
            sd = 0
            do m = 1,3
               taup_head(m) = d_points(2,m)
               taup_tail(m) = d_points(1,m)
               taud_head(m) = p_points(m)

               taup(m) = taup_head(m) - taup_tail(m)
               taud(m) = taud_head(m) - taud_tail(m)

               sp = sp + taup(m)*taup(m)
               sd = sd + taud(m)*taud(m)
            enddo
            sp = sqrt(sp)
            sd = sqrt(sd)

            ca = (taup(1)*taud(1) + taup(2)*taud(2) +
     &            taup(3)*taud(3))/(sp*sd)
            sa = sqrt(1.d0 - ca**2)
            cta = ca/sa

            a1 = 1.d0/sa
            a2 = -cta

c            pcval(k,(1-i)/2+1) = (sp/sd)*a1/lenpi(k)
c            ucval(k,(1-i)/2+1) = a2/lenpi(k)

c           # lenpi(k) = sp
c           # pcval is needed for computing rhs...
            pcval(k,(1-i)/2+1) = a1/sd

            p((1-i)/2+1) = a1/sd
            u((1-i)/2+1) = a2/sp
         enddo
         tv(k) = (u(1) + u(2))/(p(1) + p(2))
      enddo
      tv(0) = tv(np)

  100 format(I5,3E24.16)
      end




c     # -------------------------------------------------------------
c     # Miscellaneous routines for allocating memory
c     # -------------------------------------------------------------

      subroutine get_tri_ptrs(mwork,mx,my,pcval,
     &      trisys_du,trisys_d,trisys_dl,trisys_du2,qh,
     &      coeffs_alpha,xpi,ypi,zpi,xdi,ydi,zdi,tv,
     &      pval,dval,rv,rhs)
      implicit none

      integer mwork,mx,my
      integer trisys_du,trisys_d,trisys_dl,trisys_du2,qh
      integer pcval,coeffs_alpha
      integer xpi,ypi,zpi,xdi,ydi,zdi,tv,pval,dval,rv,rhs

      integer total, nbc

      nbc = 2*(mx/2 + my)

      pcval        = 1
      trisys_du    = pcval        + 2*nbc
      trisys_d     = trisys_du    + nbc
      trisys_dl    = trisys_d     + nbc
      trisys_du2   = trisys_dl    + nbc
      qh           = trisys_du2   + nbc
      coeffs_alpha = qh           + nbc
      xpi          = coeffs_alpha + 3
      ypi          = xpi          + 3*nbc
      zpi          = ypi          + 3*nbc
      xdi          = zpi          + 3*nbc
      ydi          = xdi          + nbc
      zdi          = ydi          + nbc
      tv           = zdi          + nbc
      pval         = tv           + nbc + 1
      dval         = pval         + 2*nbc
      rv           = dval         + nbc
      rhs          = rv           + nbc + 1
      total        = rhs          + nbc + 1

      total = total - 1

      if (mwork .lt. 0) then
c        # This is being called from the RKC solver - we already
c        # know that we have enough space
         return
      endif
      if (total .gt. mwork) then
         write(6,*) 'Increase memory in work_tri to ',total
         stop
      endif

      end
