c     # -------------------------------------------------------------
c     # Function describing right hands sides for explicit ODE solvers
c     # (Forward Euler, Rock4, RKC,IRKC) and matrix-vector multiplies for
c     # implicit solvers (Backward Euler, Crank-Nicolson, TR-BDF2.
c     #
c     # Functions in this file should be used for domains with physical
c     # boundary conditions.  For the sphere, see 'func_sphere'.
c     # -------------------------------------------------------------


c     # --------------------------------------------------------------
c     # Right-hand side for RKC, ROCK4, and Forward Euler
c     # --------------------------------------------------------------
      subroutine f_exp(neqn,t,q,rhs)
      implicit none

      integer neqn
      double precision t, q(neqn), rhs(neqn)

      integer mx,my,c_bc,meqn, ldbc, ldtri, qd
      integer mbc
      double precision xlower,ylower,dx,dy

      include "para.i"

      mx = mx_com
      my = my_com

      meqn = meqn_com
      mbc = mbc_com

c     # Compute dual cell values (user defined)
      call compute_dual_sphere(mx,my,meqn,q,work(qd_ptr))

      call apply_equator_bc(mx,my,meqn,q,work(qd_ptr))

c     # Compute Laplacian using primal and dual values
c     # (user defined)
      call apply_LB(mx,my,mbc,meqn,work(coeffs_ptr),work(area_ptr),q,
     &      work(qd_ptr),work(diff_ptr))

      call src_reaction(mx,my,meqn,t,q,work(reac_ptr))
      call src_diffusion(mx,my,meqn,t,q,work(diff_ptr))
      call compute_rhs_exp(mx,my,meqn,t,work(diff_ptr),
     &      work(reac_ptr),rhs)

c     # Zero out rhs ghost cell values so RKC won't update them
      call set_zeros(mx,my,meqn,rhs)

      end

      subroutine compute_rhs_exp(mx,my,meqn,t,diff,reac,rhs)
      implicit none

      integer mx,my,meqn
      double precision t
      double precision diff(0:mx+1,0:my+1,*)
      double precision reac(0:mx+1,0:my+1,*)
      double precision rhs(0:mx+1,0:my+1,*)

      integer i,j,mq,q_idx,m
      logical active_parabolic_term

      do mq = 1,meqn
         if (active_parabolic_term(mq)) then
            m = q_idx(mq)
            do j = 1,my
               do i = 1,mx
                  rhs(i,j,m) = diff(i,j,m) + reac(i,j,m)
               enddo
            enddo
         endif
      enddo

      end

c     # --------------------------------------------------------------
c     # Right hand side for IRKC.
c     # This is also an explicit solver, but has different input arguments
c     # --------------------------------------------------------------
      subroutine f_exp_irkc(neqn,t,q,rhs)
      implicit none

      integer neqn
      double precision t, q(neqn), rhs(neqn)

      integer mx,my,meqn,mbc, mpara

      integer m

      include "para.i"

      mx = mx_com
      my = my_com

      meqn = meqn_com
      mpara = mpara_com
      mbc = mbc_com

      call copy_to_work_irkc(mx,my,meqn,mpara,q,work(q_ptr))
      call copy_to_work_irkc(mx,my,meqn,mpara,rhs,work(rhs_ptr))

      call compute_dual_sphere(mx,my,meqn,work(q_ptr),work(qd_ptr))

      call apply_equator_bc(mx,my,meqn,work(q_ptr),work(qd_ptr))

      call apply_LB(mx,my,mbc,meqn,work(coeffs_ptr),work(area_ptr),
     &      work(q_ptr),work(qd_ptr),work(rhs_ptr))

      call src_diffusion(mx,my,meqn,t,work(q_ptr),work(rhs_ptr))

      call set_zeros(mx,my,meqn,work(rhs_ptr))

      call copy_from_work_irkc(mx,my,meqn,mpara,work(rhs_ptr),rhs)

      end


      subroutine f_imp_irkc(ieqn,nactive,t,q,rhs,compute_jac,jac)
      implicit none

      integer ieqn, nactive
      double precision t, q(nactive), rhs(nactive), jac(nactive,nactive)
      logical compute_jac

      integer mx,my, meqn

      integer i,j,m,n, ii,jj

      include "para.i"

      mx = mx_com
      my = my_com
      meqn = meqn_com

c      r1 = MOD(ieqn-1,mx+2) + 1;
c      j = (ieqn - r1)/(mx+2) + 1;
c      i = ieqn - (j-1)*(mx+2);

c     # For ieqn = 1,P+1,2*P+1, 3*P+1, where P = nactive  = number of PDEs
c     #
c     # ieqn = (N-1)*P + 1
c     # N = (mx+2-1)*j + i

      n = (ieqn-1)/nactive + 1

      i = mod(n-1,mx+2) + 1
      j = (n - i)/(mx+2) + 1

c     # Subtract 1, since we grid points start at (0,0)
      i = i - 1
      j = j - 1

      if (i .eq. 0 .OR. i .eq. mx+1 .or. j .eq. 0 .or. j .eq. my+1) then
         do m = 1,nactive
            rhs(m) = 0
         enddo
         do ii = 1,nactive
            do jj = 1,nactive
               jac(ii,jj) = 0.d0
            enddo
         enddo
         return
      endif

      call src_reaction_irkc(i,j,nactive,t,q,rhs,compute_jac,jac)

      end

c     # --------------------------------------------------------------
c     # Crank-Nicolson
c     # --------------------------------------------------------------
      subroutine compute_rhs_cn(mx,my,meqn,qn,t,dt,diff,reac,rhs)
      implicit none

      integer mx,my,meqn
      double precision t,dt
      double precision qn(0:mx+1,0:my+1,*)
      double precision rhs(0:mx+1,0:my+1,*)
      double precision diff(0:mx+1,0:my+1,*)
      double precision reac(0:mx+1,0:my+1,*)

      integer i,j,m, mq, q_idx, mbc
      logical active_parabolic_term
      double precision xlower,ylower, dx,dy
      double precision lambda, tp
      integer ldbc,ldtri

      include 'para.i'

      mbc = mbc_com

c     # Compute dual cell values (user defined)
      call compute_dual_sphere(mx,my,meqn,qn,work(qd_ptr))

      call apply_equator_bc(mx,my,meqn,qn,work(qd_ptr))

c     # Compute Laplacian using primal and dual values
c     # (user defined)
      call apply_LB(mx,my,mbc,meqn,work(coeffs_ptr),
     &      work(area_ptr),qn,work(qd_ptr),diff)

c     # Time center the source term for Crank-Nicolson
      tp = t + dt/2.d0
      call src_reaction(mx,my,meqn,tp,qn,reac)

      lambda = -2.d0/dt
      do mq = 1,meqn
         if (active_parabolic_term(mq)) then
            m = q_idx(mq)
            do j = 1,my
               do i = 1,mx
                  rhs(i,j,m) = lambda*qn(i,j,m) - diff(i,j,m) -
     &                  2.d0*reac(i,j,m)
               enddo
            enddo
         endif
      enddo

      end

c     # --------------------------------------------------------------
c     # Matrix vector multiply for CN, TR-BDF2 and BE.
c     # --------------------------------------------------------------
      subroutine matvec(mx,my,meqn,q,lambda,Aq,ishomogeneous)
      implicit none

      integer mx,my,meqn
      double precision q(0:mx+1,0:my+1,*)
      double precision Aq(0:mx+1,0:my+1,*)
      double precision lambda
      logical ishomogeneous

      integer ldbc,ldtri
      double precision xlower,ylower,dx,dy
      double precision t_curr, dt_curr, tp

      integer i,j,m,mq,q_idx, mbc
      logical active_parabolic_term

      include 'para.i'

      mbc = mbc_com

c     # Compute dual cell values (user defined)
      call compute_dual_sphere(mx,my,meqn,q,work(qd_ptr))

      call apply_equator_bc(mx,my,meqn,q,work(qd_ptr))

c     # Compute Laplacian using primal and dual values
c     # (user defined)
      call apply_LB(mx,my,mbc,meqn,work(coeffs_ptr),work(area_ptr),q,
     &      work(qd_ptr),Aq)

      do mq = 1,meqn
         if (active_parabolic_term(mq)) then
            m = q_idx(mq)
            do j = 1,my
               do i = 1,mx
                  Aq(i,j,m) = Aq(i,j,m) + lambda*q(i,j,m)
               enddo
            enddo
         endif
      enddo

      call set_zeros(mx,my,meqn,Aq)

      end
