c     # Compute an approximate unit normal to cell edge
c     #
c     # Inputs :
c     # taud     Vector between adjoining cell centers ("dual" cell edge)
c     # taup     Vector between edge nodes ("primal" cell edge)
c     #
c     # Ouputs :
c     # nv       Unit vector vector normal to taup, and tangent to the surface
c     # sp       Length of taup
c     #
c     # Idea is to construct normal to edge vector using
c     #
c     #     t^i = a^{i1} t_1 + a^{i2} t_2
c     #
c     # where t_1, t_2 are coordinate basis vectors T_xi, T_eta
c     # and t^i dot t_j = 1 (i == j), 0 (i /= j)
c     #
c     # At left edge, we have
c     #       taud ~ T_xi = t_1  and   taup ~ T_eta = t_2
c     #
c     # At bottom edge, we have
c     #       taup ~ T_xi = t_1  and   taud ~ T_eta = t_2
c     #
c     # Metric a_{ij} = t_i dot t_j
c     # Metric a^{ij} is the inverse of a_{ij}
c     #
c     # Both versions use this idea.  That is, both compute
c     # coefficients c1, c2 so that the unit normal at the edge
c     # is expressed as
c     #
c     #        n = c1*taup + c2*taud
c     #
c     # In version 1, trig. identities are used to express these coefficients
c     # in terms of the angle between taup and taud.  In version 2, entries of
c     # the conjugate tensor are used directly.   Both compute the same vector, but
c     # version 2 is about 20%-30% faster.
c     #
c     # Formulas work for both left and bottom edges.
c     #
c     # Neither version depends on any knowledge of the surface normals.

      subroutine get_normal(taup,taud,nv,sp)
      implicit none

      double precision taup(3),taud(3),nv(3), sp
      integer version

      data version /2/

      if (version .eq .1) then
         call get_normal_ver1(taup,taud,nv,sp)
      elseif (version .eq. 2) then
         call get_normal_ver2(taup,taud,nv,sp)
      endif

      end

      subroutine get_normal_ver1(taup,taud,nv,sp)
      implicit none

      double precision taup(3),taud(3),nv(3)
      double precision sp,sd,dt,ct,st,c1,c2
      double precision tp(3), td(3)

      integer m

      sp = 0
      sd = 0
      do m = 1,3
         sp = sp + taup(m)*taup(m)
         sd = sd + taud(m)*taud(m)
      enddo
      sp = sqrt(sp)
      sd = sqrt(sd)

c     # st = sin(theta)
c     # ct = cos(theta)
c     # theta = angle between taup and taud
      ct = 0
      do m = 1,3
         tp(m) = taup(m)/sp
         td(m) = taud(m)/sd
         ct = ct + tp(m)*td(m)
      enddo
      st = sqrt(1.d0-ct*ct)

      c1 = 1.d0/st
      c2 = -ct/st
      do m = 1,3
         nv(m) = c1*td(m) + c2*tp(m)
      enddo
      end

      subroutine get_normal_ver2(taup,taud,nv,sp)
      implicit none

      double precision taup(3),taud(3),nv(3), sp

      integer m
      double precision nlen,aii,ai2,c1,c2

c     # we leave out any scaling by the determinant of the metric,
c     # since we just normalize the vector anyway.
      aii = 0
      ai2 = 0
      do m = 1,3
         aii = aii + taup(m)*taup(m)
         ai2 = ai2 - taud(m)*taup(m)
      enddo

      nlen = 0
      do m = 1,3
         nv(m) = aii*taud(m) + ai2*taup(m)
         nlen = nlen + nv(m)*nv(m)
      enddo
      nlen = sqrt(nlen)

c      c1 = aii/nlen
c      c2 = ai2/nlen

      do m = 1,3
c         nv(m) = c1*taud(m) + c2*taup(m)
         nv(m) = nv(m)/nlen
      enddo
      sp = sqrt(aii)


      end
