c     # -----------------------------------------------------
c     # setup_parabolic_phys
c     #
c     # Setup parabolic solver for problems with physical
c     # boundary conditions.
c     #
c     # Boundary coefficients are set in routine 'assign_bc',
c     # which will be called from this setup procedure.
c     # -----------------------------------------------------

      subroutine setup_parabolic(mx,my,meqn,mbc,plist,method,
     &      xlower,ylower,dx,dy)
      implicit none

      integer mx,my,meqn, method,mbc
      logical plist(meqn)

      double precision xlower,ylower,dx,dy

c     # Set up pointers in common blocks in 'para.i'
      call setup_common_phys(mx,my,mbc,xlower,ylower,dx,dy,meqn,
     &      plist,method)

c     # Set up mesh data, including areas of mesh cells.
      call setup_parabolic_geo(mx,my,mbc,xlower,ylower,dx,dy)

c     # Set up boundary conditions for physical boundaries.
      call setup_parabolic_bc(mx,my,meqn,xlower,ylower,dx,dy)

      end

      subroutine setup_parabolic_geo(mx,my,mbc,xlower,ylower,dx,dy)
      implicit none

      integer mx,my,mbc
      double precision xlower,ylower,dx,dy

      include "para.i"

c     # Set up mesh.  Use pointers defined in para.i
      call setup_mesh(mx,my,mbc,xlower,ylower,dx,dy,
     &      work(xp_ptr),work(yp_ptr),work(zp_ptr),
     &      work(xd_ptr),work(yd_ptr),work(zd_ptr))


c     # Set up approximate metric terms (lengths of boundary edges; mesh
c     # cell areas)
      call compute_area(mx,my,mbc,
     &      work(xd_ptr),work(yd_ptr),work(zd_ptr),
     &      work(area_ptr))

      call compute_coeffs(mx,my,mbc,
     &      work(xp_ptr),work(yp_ptr),work(zp_ptr),
     &      work(xd_ptr),work(yd_ptr),work(zd_ptr),
     &      work(coeffs_ptr))

      call setup_boundary_mesh(mx,my,mbc,xlower,ylower,dx,dy,
     &      work(xp_ptr),work(yp_ptr),work(zp_ptr),
     &      work(xd_ptr),work(yd_ptr),work(zd_ptr),
     &      work(xpi_ptr),work(ypi_ptr),work(zpi_ptr),
     &      work(xdi_ptr),work(ydi_ptr),work(zdi_ptr))

      call set_bc_coeffs(mx,my,
     &      work(xpi_ptr),work(ypi_ptr),work(zpi_ptr),
     &      work(xdi_ptr),work(ydi_ptr),work(zdi_ptr),
     &      work(coeffs_ptr),work(len_ptr))

      end

      subroutine setup_parabolic_bc(mx,my,meqn,xlower,ylower,dx,dy)
      implicit none

      integer mx,my, meqn
      double precision xlower,ylower,dx,dy

      integer ldtri, ldbc

      include "para.i"

      ldtri = ldtri_com
      ldbc = ldbc_com

c     # Called user-defined function which assigns
c     # boundary conditions.
      call assign_bc(mx,my,meqn,xlower,ylower,dx,dy,
     &      work(a_bc_ptr),work(b_bc_ptr),work(c_bc_ptr),ldbc)

c     # Set up tridiagonal system for solving boundary conditions.
      call initialize_bc(mx,my,meqn,work(coeffs_ptr),
     &      work(a_bc_ptr),work(b_bc_ptr),work(c_bc_ptr),
     &      ldbc,work(len_ptr),work(work_tri_ptr),ldtri, ipiv)

      end

      subroutine setup_common_phys(mx,my,mbc,xlower,ylower,dx,dy,
     &      meqn,plist,method)
      implicit none

      integer mx,my, meqn, method,mbc
      logical plist(meqn)

      double precision xlower,ylower,dx,dy
      integer np, nd, nbc, total, t

      integer i, m, mq, mpara, ldtri, ldbc
      integer get_rkc_mem, get_trbdf_mem, get_rock4_mem
      integer get_cn_mem, get_beuler_mem, get_feuler_mem
      integer get_irkc_mem
      integer mwork_solver, method_mem(7)

      integer np1, nd1, mbc1

      include "para.i"

      nfevals_total = 0

      if (meqn .gt. maxmeqn) then
         write(6,*) 'para.i : Increase size of maxmeqn to ',meqn
         stop
      endif

      m = 0
      do mq = 1,meqn
         plist_com(mq) = plist(mq)
         if (plist(mq)) then
            m = m + 1
         endif
      enddo
      mpara = m

c     # To communicate these through common blocks
      mx_com = mx
      my_com = my
      dx_com = dx
      dy_com = dy
      xlower_com = xlower
      ylower_com = ylower

      np   = (mx+2)*(my+2)
      nd   = (mx+1)*(my+1)
      nbc  = 2*(mx + my)
      ldtri = 15*nbc + 6
      ldbc = max(mx,my)

      mpara_com = mpara
      meqn_com = meqn
      mbc_com = mbc
      ldtri_com = ldtri
      ldbc_com = ldbc

c     # Needed for BiCG solver
      method_mem(1) = get_feuler_mem(mx,my,mpara)
      method_mem(2) = get_beuler_mem(mx,my,mpara)
      method_mem(3) = get_cn_mem(mx,my,mpara)
      method_mem(4) = get_trbdf_mem(mx,my,mpara)
      method_mem(5) = get_rock4_mem(mx,my,mpara)
      method_mem(6) = get_rkc_mem(mx,my,mpara)
      method_mem(7) = get_irkc_mem(mx,my,mpara)
      mwork_solver  = method_mem(method)

c     # We allow for an extra row of ghost cells, which
c     # may be overkill, but well, this is what I do...
      mbc1 = mbc + 1
      np1 = (mx+2*mbc1)*(my+2*mbc1)
      nd1 = (mx+2*mbc1 + 1)*(my+2*mbc1+1)


c     # Pointers into work array
      xp_ptr           = 1
      yp_ptr           = xp_ptr           + np1
      zp_ptr           = yp_ptr           + np1
      area_ptr         = zp_ptr           + np1
      coeffs_ptr       = area_ptr         + np1
      len_ptr          = coeffs_ptr       + 4*np
      xd_ptr           = len_ptr          + nbc
      yd_ptr           = xd_ptr           + nd1
      zd_ptr           = yd_ptr           + nd1

      xpi_ptr          = zd_ptr           + nd1
      ypi_ptr          = xpi_ptr          + 3*nbc
      zpi_ptr          = ypi_ptr          + 3*nbc

      xdi_ptr          = zpi_ptr          + 3*nbc
      ydi_ptr          = xdi_ptr          + nbc
      zdi_ptr          = ydi_ptr          + nbc

      normals_ptr      = zdi_ptr          + nbc

      a_bc_ptr         = normals_ptr      + 6*nd1
      b_bc_ptr         = a_bc_ptr         + 4*ldbc*meqn
      c_bc_ptr         = b_bc_ptr         + 4*ldbc*meqn
      work_tri_ptr     = c_bc_ptr         + 4*ldbc*meqn
      qd_ptr           = work_tri_ptr     + ldtri*meqn
      rhs_ptr          = qd_ptr           + nd*mpara
      q_ptr            = rhs_ptr          + np*mpara
      diff_ptr         = q_ptr            + np*mpara
      reac_ptr         = diff_ptr         + np*mpara
      work_solver_ptr  = reac_ptr         + np*mpara
      total            = work_solver_ptr  + mwork_solver

      total = total - 1

      if (mwork .lt. total) then
         write(6,100) 'para.i : Increase memory ',
     &         'in work array to at least ', total
         stop
      endif
  100 format(A,A,I10)

c     # Check integer array 'ipiv'.  There is only one
c     # array, so we don't need a pointer into it.
      if (miwork .lt. 2*(mx + my)*meqn) then
         write(6,100) 'para.i : Increase size of ipiv ',
     &         'to ', 2*(mx + my)*meqn
         stop
      endif

      end


      subroutine set_homogeneous_bc(mx,my,meqn,c_bc,ldbc)
      implicit none

      integer mx,my,ldbc, meqn
      double precision c_bc(ldbc,4,meqn)

      integer i,j, mq

c     # meqn is just 1, so we don't really use it here..
      do mq = 1,meqn
         do i = 1,mx
            c_bc(i,1,mq) = 0
            c_bc(i,3,mq) = 0
         enddo

         do j = 1,my
            c_bc(j,2,mq) = 0
            c_bc(j,4,mq) = 0
         enddo
      enddo
      end


      subroutine set_bc_coeffs(mx,my,xpi,ypi,zpi,xdi,ydi,zdi,
     &      coeffs,lenp)
      implicit none

      integer mx,my
      double precision xpi(2*(mx+my),-1:1)
      double precision ypi(2*(mx+my),-1:1)
      double precision zpi(2*(mx+my),-1:1)

      double precision xdi(2*(mx+my))
      double precision ydi(2*(mx+my))
      double precision zdi(2*(mx+my))

      double precision coeffs(0:mx+1,0:my+1,2,2)
      double precision lenp(2*(mx+my))

      double precision taup(3), taud(3)
      double precision c1,c2, sp

      integer i,j,k, kp1, m

      k = 1
      do i = 1,mx
         taud(1) = xpi(k,1) - xpi(k,0)
         taud(2) = ypi(k,1) - ypi(k,0)
         taud(3) = zpi(k,1) - zpi(k,0)

         taup(1) = xdi(k+1) - xdi(k)
         taup(2) = ydi(k+1) - ydi(k)
         taup(3) = zdi(k+1) - zdi(k)

         call get_coeffs(taup,taud,c1,c2,sp)
         coeffs(i,1,2,1) = c1
         coeffs(i,1,2,2) = c2

         lenp(k) = sp

         k = k + 1
      enddo

      do j = 1,my
         taud(1) = xpi(k,0) - xpi(k,1)
         taud(2) = ypi(k,0) - ypi(k,1)
         taud(3) = zpi(k,0) - zpi(k,1)

         taup(1) = xdi(k+1) - xdi(k)
         taup(2) = ydi(k+1) - ydi(k)
         taup(3) = zdi(k+1) - zdi(k)

c        # sp is the length of taup
         call get_coeffs(taup,taud,c1,c2,sp)

         coeffs(mx+1,j,1,1) = c1
         coeffs(mx+1,j,1,2) = c2

         lenp(k) = sp

         k = k + 1
      enddo

      do i = mx,1,-1
         taud(1) = xpi(k,0) - xpi(k,1)
         taud(2) = ypi(k,0) - ypi(k,1)
         taud(3) = zpi(k,0) - zpi(k,1)

         taup(1) = xdi(k) - xdi(k+1)
         taup(2) = ydi(k) - ydi(k+1)
         taup(3) = zdi(k) - zdi(k+1)

         call get_coeffs(taup,taud,c1,c2,sp)

         coeffs(i,my+1,2,1) = c1
         coeffs(i,my+1,2,2) = c2

         lenp(k) = sp

         k = k + 1
      enddo

      do j = my,1,-1
         taud(1) = xpi(k,1) - xpi(k,0)
         taud(2) = ypi(k,1) - ypi(k,0)
         taud(3) = zpi(k,1) - zpi(k,0)

         if (j .eq. 1) then
            kp1 = 1
         else
            kp1 = k+1
         endif
         taup(1) = xdi(k) - xdi(kp1)
         taup(2) = ydi(k) - ydi(kp1)
         taup(3) = zdi(k) - zdi(kp1)

c        # sp is the length of taup
         call get_coeffs(taup,taud,c1,c2,sp)
         coeffs(1,j,1,1) = c1
         coeffs(1,j,1,2) = c2

         lenp(k)   = sp

         k = k + 1
      enddo

      end


c     # -------------------------------------------------------
c     # Apply the boundary conditions for the Cartesian grid
c     #
c     # In this case, BCs can be applied in the usual way.  We
c     # Don't need to solve a tridiagonal system.
c     # ----------------------------------------------------------
      subroutine apply_cart_bc(mx,my,meqn,a_bc,b_bc,
     &      c_bc,ldbc,t,q)
      implicit none

      integer mx,my, meqn, ldbc
      double precision t
      double precision q(0:mx+1,0:my+1,meqn)
      double precision a_bc(ldbc,4,meqn)
      double precision b_bc(ldbc,4,meqn)
      double precision c_bc(ldbc,4,meqn)

      logical is_periodic

      integer i,j, m

      if (is_periodic()) then
         do m = 1,meqn
c           # Apply periodic boundary conditions
            do j = 1,my
               q(0,j,m) = q(mx,j,m)
               q(mx+1,j,m) = q(1,j,m)
            enddo

            do i = 0,mx+1
               q(i,0,m) = q(i,my,m)
               q(i,my+1,m) = q(i,1,m)
            enddo
         enddo
      else
c        # Need to add something here!
         write(6,*) 'apply_cart_bc (parabolic_all.f) : Need to add ',
     &         'Cartesian BCs for general case'
         stop
      endif


      end

      subroutine set_periodic()
      implicit none

      logical isperiodic_com
      common /combc_cart/ isperiodic_com

      isperiodic_com = .true.

      end

      logical function is_periodic()
      implicit none

      logical isperiodic_com
      common /combc_cart/ isperiodic_com

      is_periodic = isperiodic_com

      end

c     # -------------------------------------------------------
c     # Apply the boundary conditions
c     #
C     # This assumes that we have physical boundary conditions on all
c     # four edges of Cartesian grid.
c     #
c     # This requires that the tridiagonal system has been
c     # set up by 'initialize_bc(...)'
c     # ----------------------------------------------------------

      subroutine apply_bc(mx,my,meqn,qp,qd,t,ishomogeneous)
      implicit none

      integer mx,my,meqn
      double precision t
      logical ishomogeneous

      double precision qp(0:mx+1,0:my+1,*)
      double precision qd(mx+1,my+1,*)

      logical iscart
      integer ldbc, ldtri
      double precision xlower, ylower, dx, dy

      include 'para.i'

      ldbc = ldbc_com
      ldtri = ldtri_com
      xlower = xlower_com
      ylower = ylower_com
      dx = dx_com
      dy = dy_com

c     # Update inhomogenous boundary data 'c_bc'.  This can
c     # be a dummy routine if 'c_bc' does not depend on time.
c     # (user defined)
      if (ishomogeneous) then
         call set_homogeneous_bc(mx,my,meqn,work(c_bc_ptr),ldbc)
      else
         call update_bc(mx,my,meqn,xlower,ylower,dx,dy,t,
     &         work(c_bc_ptr),ldbc)
      endif

      if (iscart()) then
         call apply_cart_bc(mx,my,meqn,work(a_bc_ptr),
     &         work(b_bc_ptr),work(c_bc_ptr),ldbc,t,qp)
      else
c        # Apply boundary conditions; both primal and dual
c        # boundary values will get set by this routine.
         call apply_general_bc(mx,my,meqn,work(work_tri_ptr),ldtri,ipiv,
     &         work(c_bc_ptr),ldbc,qp,qd)
      endif

      end


      subroutine apply_general_bc(mx,my,meqn,work_tri,ldtri,ipiv,c_bc,
     &      ldbc,qp,qd)
      implicit none

      integer mx,my,meqn,ldtri,ldbc, mpara
      double precision c_bc(ldbc,4,meqn)

c     # The last dimension is the number of equations
c     # containing parabolic terms; this is <= meqn
      double precision qp(0:mx+1,0:my+1,*)
      double precision qd(mx+1,my+1,*)
      double precision work_tri(ldtri,meqn)
      integer ipiv(2*(mx+my),meqn)

      integer j

      integer m,mq
      logical active_parabolic_term
      integer q_idx

      do mq = 1,meqn
         if (active_parabolic_term(mq)) then
            m = q_idx(mq)
            call apply_bc_internal(mx,my,work_tri(1,mq),ldtri,
     &            ipiv(1,mq),ldbc,c_bc(1,1,mq),qp(0,0,m),qd(1,1,m))
         endif
      enddo
      end

      subroutine apply_bc_internal(mx,my,work_tri,mwork_tri,
     &      ipiv,ldbc,c_bc,qp,qd)
      implicit none

      integer mx,my,mwork_tri,miwork_tri, ldbc
      double precision qp(0:mx+1,0:my+1)
      double precision qd(mx+1,my+1)
      double precision work_tri(mwork_tri)
      double precision ipiv(2*(mx+my))

      double precision c_bc(ldbc,4)

      double precision fa

      integer pcval,ucval,bcval
      integer trisys_du, trisys_d, trisys_dl, trisys_du2, qh
      integer coeffs_alpha, dval, pval, rhs, rv,tv

      integer ntri, nrhs, ldb, info
      character trans

      integer nbc

      nbc = 2*(mx + my)

c      mwork_tri = 15*nbc + 6
      call get_tri_ptrs(mwork_tri,mx,my,pcval,ucval,bcval,
     &      trisys_du,trisys_d,trisys_dl,trisys_du2,qh,
     &      coeffs_alpha,pval,dval,rhs,rv,tv)

      call update_vectors(mx,my,ldbc,c_bc,qp,work_tri(bcval),
     &      work_tri(pval))

c     # Get time dependent right hand side and scalar fa
      call get_tri_rhs(mx,my,
     &      work_tri(pcval),work_tri(ucval),work_tri(bcval),
     &      work_tri(pval),work_tri(rhs),work_tri(rv),fa)

c     # Solve tridiagonal system to get dual bc values.
      trans = 'n'
      nrhs = 1
      ntri = nbc - 1
      ldb = ntri
      call dgttrs(trans, ntri,nrhs,work_tri(trisys_dl),
     &      work_tri(trisys_d),work_tri(trisys_du),
     &      work_tri(trisys_du2),ipiv,
     &      work_tri(rhs),ldb,info)

c     # Assign primal values
      call assign_bc_tri(mx,my,
     &      work_tri(pcval),work_tri(ucval),work_tri(bcval),
     &      work_tri(pval),work_tri(dval),
     &      work_tri(qh),work_tri(rhs),
     &      work_tri(coeffs_alpha),fa,qp,qd)
      end


c     # -------------------------------------------------------
c     # Compute the right hand side based on current solution
c     # -------------------------------------------------------
      subroutine get_tri_rhs(mx,my,pcval,ucval,bcval,
     &      pval,rhs,rv,fa)
      implicit none

      integer mx,my

      double precision pcval(2*(mx+my))
      double precision ucval(2*(mx+my))
      double precision bcval(2*(mx+my),3)

      double precision rhs(2*(mx+my))
      double precision pval(2*(mx+my))
      double precision fa

      double precision a,b,c,p

      double precision rv(0:2*(mx + my))

      integer k, nbc

      nbc = 2*(mx + my)

      do k = 1,nbc
         a = bcval(k,1)
         b = bcval(k,2)
         c = bcval(k,3)
         p = pcval(k)
c        # Eqn (3.47)
         rv(k) = (c + b*p*pval(k))/(a + b*p)
      enddo
  120 format(I5,2E16.8)
      rv(0) = rv(nbc)

      do k = 1,nbc
c        # Eqn (3.51)
         rhs(k) = -(rv(k-1) + rv(k))
      enddo
c     # Eqn (3.51)
      fa = rhs(nbc)

      end


      subroutine update_vectors(mx,my,ldbc,c_bc,qp,bcval,qval)
      implicit none

      integer mx,my, ldbc
      double precision c_bc(ldbc,4)
      double precision qp(0:mx+1,0:my+1)

      double precision bcval(2*(mx+my),3)
      double precision qval(2*(mx+my))

      integer i, j, n, nbc

      nbc = 2*(mx + my)

c     # Eqn (3.37)
      n = 1
      do i = 1,mx
         qval(n) = qp(i,1)
         bcval(n,3) = c_bc(i,1)
         n = n + 1
      enddo

      do j = 1,my
         qval(n) = qp(mx,j)
         bcval(n,3) = c_bc(j,2)
         n = n + 1
      enddo

      do i = mx,1,-1
         qval(n) = qp(i,my)
         bcval(n,3) = c_bc(i,3)
         n = n + 1
      enddo

      do j = my,1,-1
         qval(n) = qp(1,j)
         bcval(n,3) = c_bc(j,4)
         n = n + 1
      enddo

      end


      subroutine assign_bc_tri(mx,my,pcval,ucval,
     &      bcval,pval,dval,qh,qinh,coeffs_alpha,fa,qp,qd)
      implicit none

      integer mx,my
      double precision qh(2*(mx+my)-1)
      double precision qinh(2*(mx+my)-1)
      double precision coeffs_alpha(3), fa

      double precision qp(0:mx+1,0:my+1)
      double precision qd(mx+1,my+1)

      double precision pcval(2*(mx+my))
      double precision ucval(2*(mx+my))
      double precision bcval(2*(mx+my),3)
      double precision dval(2*(mx+my)+1)
      double precision pval(2*(mx+my))

      integer i,j,n, nbc
      double precision alpha, a,b,c

      nbc = 2*(mx + my)

c     # Eqn. (3.46)
      alpha = coeffs_alpha(1)*fa + coeffs_alpha(2)*qinh(1) +
     &      coeffs_alpha(3)*qinh(nbc-1)

      if (1./alpha .eq. 0) then
         alpha = 0
      endif

c     # Eqn. (3.44)
      do n = 1,nbc-1
         dval(n) = qinh(n) + alpha*qh(n)
      enddo
      dval(nbc) = alpha
      dval(nbc+1) = dval(1)


c     # Overwrite pval...
      do n = 1,nbc
         a = bcval(n,1)
         b = bcval(n,2)
         c = bcval(n,3)
c        # Solve Eqn. (3.43) for primal boundary values (qkp)
         pval(n) = (c + b*(pcval(n)*pval(n) -
     &         ucval(n)*(dval(n+1)-dval(n))))/(a + b*pcval(n))
      enddo

c     # Set primal boundary values.  Dual values will be averaged
      n = 1
      do i = 1,mx
         qp(i,0) = pval(n)
         qd(i,1) = dval(n)
         n = n + 1
      enddo

      do j = 1,my
         qp(mx+1,j) = pval(n)
         qd(mx+1,j) = dval(n)
         n = n + 1
      enddo

      do i = mx,1,-1
         qp(i,my+1) = pval(n)
         qd(i+1,my+1) = dval(n)
         n = n + 1
      enddo

      do j = my,1,-1
         qp(0,j) = pval(n)
         qd(1,j+1) = dval(n)
         n = n + 1
      enddo

      end


c     # -------------------------------------------------------
c     # Initialize tridiagonal systems needed for applying BCs
c     #
c     # Assume BCs of the form
c     #
c     #        A q + B q_n = C(t)
c     #
c     # where A, B, and C depend on space.  Only C can depend on
c     # time.
c     #
c     # This assumes that we have already set up 'coeffs_p', and
c     # the user has supplied BC coefficients A and B in arrays
c     # a_bc, b_bc.  The user also supplies C at time t = 0.
c     # If C is time dependent, the user can update the values
c     # during calls from the RKC solver.
c     #
c     # This constructs a tridiagonal system each equation
c     # containing boundary terms.  This means that different
c     # species can have different type boundary conditions.
c     # ----------------------------------------------------------
      subroutine initialize_bc(mx,my,meqn,coeffs_p,
     &      a_bc,b_bc,c_bc,ldbc,lenp,work_tri,ldtri,ipiv)
      implicit none

      integer mx,my, meqn, ldbc,ldtri

      double precision coeffs_p(0:mx+1,0:my+1,2,2)

      double precision a_bc(ldbc,4,meqn)
      double precision b_bc(ldbc,4,meqn)
      double precision c_bc(ldbc,4,meqn)

      double precision lenp(2*(mx+my))
      double precision work_tri(ldtri,meqn)
      integer ipiv(2*(mx+my),meqn)

      integer mq
      logical has_parabolic_term

      do mq = 1,meqn
         if (has_parabolic_term(mq)) then
c           # Initialize tridiagonal system needed for imposing BCs
            call initialize_bc_internal(mx,my,coeffs_p,lenp,
     &            a_bc(1,1,mq),b_bc(1,1,mq),c_bc(1,1,mq),ldbc,
     &            ipiv(1,mq),work_tri(1,mq),ldtri)
         endif
      enddo

      end


      subroutine initialize_bc_internal(mx,my,coeffs_p,lenp,
     &      a_bc,b_bc,c_bc,ldbc,ipiv,work_tri,mwork_tri)
      implicit none

      integer mx,my,mwork_tri,miwork_tri,ldbc
      double precision coeffs_p(0:mx+1,0:my+1,2,2)
      double precision work_tri(mwork_tri)
      integer ipiv(*)

      double precision lenp(2*(mx+my))
      double precision a_bc(ldbc,4)
      double precision b_bc(ldbc,4)
      double precision c_bc(ldbc,4)

      integer trisys_du,trisys_d,trisys_dl,trisys_du2
      integer pcval, ucval, bcval,qh, coeffs_alpha
      integer pval,dval,rhs,rv,tv
      integer nbc

      call get_tri_ptrs(mwork_tri,mx,my,pcval,ucval,bcval,
     &      trisys_du,trisys_d,trisys_dl,trisys_du2,qh,
     &      coeffs_alpha,pval,dval,rhs,rv,tv)

      call set_vectors(mx,my,coeffs_p,a_bc,b_bc,c_bc,ldbc,lenp,
     &      work_tri(pcval),work_tri(ucval),work_tri(bcval))

      call construct_sys(mx,my,work_tri(pcval),
     &      work_tri(ucval),work_tri(bcval),work_tri(tv),
     &      work_tri(trisys_dl),work_tri(trisys_d),
     &      work_tri(trisys_du),work_tri(trisys_du2),
     &      ipiv,work_tri(qh),work_tri(coeffs_alpha))

      end


      subroutine set_vectors(mx,my,coeffs_p,a_bc,b_bc,c_bc,ldbc,
     &      lenp,pcval,ucval,bcval)
      implicit none

      integer mx,my,ldbc
      double precision coeffs_p(0:mx+1,0:my+1,2,2)

      double precision lenp(2*(mx+my))
      double precision a_bc(ldbc,4)
      double precision b_bc(ldbc,4)
      double precision c_bc(ldbc,4)

      double precision pcval(2*(mx+my))
      double precision ucval(2*(mx+my))
      double precision bcval(2*(mx+my),3)

      integer i,j,n

c     # Set vectors in Eqn. (3.39)
c     # Signs on coefficients are defined as in Eqn. (3.40) and (3.41)
      n = 1
      do i = 1,mx
         pcval(n) =  coeffs_p(i,1,2,1)/lenp(i)
         ucval(n) =  -coeffs_p(i,1,2,2)/lenp(i)
         bcval(n,1) = a_bc(i,1)
         bcval(n,2) = b_bc(i,1)
         bcval(n,3) = c_bc(i,1)
         n = n + 1
      enddo


      do j = 1,my
         pcval(n) = coeffs_p(mx+1,j,1,1)/lenp(j)
         ucval(n) = coeffs_p(mx+1,j,1,2)/lenp(j)
         bcval(n,1) = a_bc(j,2)
         bcval(n,2) = b_bc(j,2)
         bcval(n,3) = c_bc(j,2)
         n = n + 1
      enddo

      do i = mx,1,-1
         pcval(n) = coeffs_p(i,my+1,2,1)/lenp(i)
         ucval(n) = -coeffs_p(i,my+1,2,2)/lenp(i)
         bcval(n,1) = a_bc(i,3)
         bcval(n,2) = b_bc(i,3)
         bcval(n,3) = c_bc(i,3)
         n = n + 1
      enddo

      do j = my,1,-1
         pcval(n)   = coeffs_p(1,j,1,1)/lenp(j)
         ucval(n)   = coeffs_p(1,j,1,2)/lenp(j)
         bcval(n,1) = a_bc(j,4)
         bcval(n,2) = b_bc(j,4)
         bcval(n,3) = c_bc(j,4)
         n = n + 1
      enddo

      end

      subroutine construct_sys(mx,my,pcval,ucval,bcval,tv,
     &      trisys_dl,trisys_d,trisys_du,trisys_du2,
     &      ipiv, qh,coeffs_alpha)
      implicit none

      integer mx,my
      double precision   coeffs_p(0:mx+1,0:my+1,2,2)
      double precision  trisys_dl(2*(mx+my))
      double precision   trisys_d(2*(mx+my))
      double precision  trisys_du(2*(mx+my))
      double precision trisys_du2(2*(mx+my))
      integer ipiv(2*(mx+my))

      double precision     qh(2*(mx+my))
      double precision coeffs_alpha(3)

      double precision pcval(2*(mx+my))
      double precision ucval(2*(mx+my))
      double precision bcval(2*(mx+my),3)

      double precision tv(0:2*(mx + my))

      double precision a,b,u,p

      double precision aa,v(2),w(2)
      integer k, km, nbc, ntri, ldb, nrhs, info, m
      character trans

      nbc = 2*(mx+my)

      do k = 1,nbc
         a = bcval(k,1)
         b = bcval(k,2)
         u = ucval(k)
         p = pcval(k)
c        # Eqn. (3.47)
         tv(k) = b*u/(a + b*p)
      enddo
      tv(0) = tv(nbc)

      do k = 1,nbc-1
         trisys_dl(k) = tv(k)
      enddo

      do k = 1,nbc
         trisys_d(k) = tv(k) - tv(k-1) - 2.d0
      enddo

      do k = 1,nbc-1
         trisys_du(k) = -tv(k)
      enddo

c     # Define vectors w and v. We only store v(1) and v(nbc-1); the rest
c     # are zero.  Same for w.
c     # Eqn. (3.49)
      w(1) = tv(nbc)
      v(1) = -tv(nbc)
      w(2) = -tv(nbc-1)
      v(2) = tv(nbc-1)

c     # Eqn. (3.50)
      u = tv(nbc) - tv(nbc-1) - 2.d0

c     # Factor tridiagonal system.
      ntri = nbc - 1
      call dgttrf(ntri,trisys_dl, trisys_d, trisys_du,
     &      trisys_du2, ipiv,info)

      info = 0
      if (info .ne. 0) then
         write(6,'(A,I5)') 'construct_sys(1) : info .ne. 0; ',
     &         info
         stop
      endif

c     # Solve for \phi_{tid}^d in Eqn. (3.45); w = qh
      do k = 1,nbc-1
         qh(k) = 0.d0
      enddo
      qh(1)    = -w(1)
      qh(nbc-1) = -w(2)

      trans = 'n'
      ldb      = ntri
      nrhs     = 1
      call dgttrs(trans,ntri,nrhs,trisys_dl,trisys_d,
     &      trisys_du,trisys_du2,ipiv,qh,ldb,info)

      if (info .ne. 0) then
         write(6,'(A,I5)') 'construct_sys(2) : info .ne. 0; ',
     &         info
         stop
      endif

  105 format(3E16.8)
c     # Coeffients needed to compute alpha.  See Eqn. (3.46)
      coeffs_alpha(1) = 1.d0/(u + v(1)*qh(1) + v(2)*qh(nbc-1))
      coeffs_alpha(2) = -coeffs_alpha(1)*v(1)
      coeffs_alpha(3) = -coeffs_alpha(1)*v(2)

      end

c     # -------------------------------------------------------------
c     # Miscellaneous routines
c     # -------------------------------------------------------------

      subroutine get_tri_ptrs(mwork,mx,my,pcval,ucval,bcval,
     &      trisys_du,trisys_d,trisys_dl,trisys_du2,qh,
     &      coeffs_alpha,pval,dval,rhs,rv,tv)
      implicit none

      integer mwork,mx,my
      integer trisys_du,trisys_d,trisys_dl,trisys_du2,qh
      integer pcval,ucval,bcval,coeffs_alpha
      integer pval,dval,rhs,rv,tv

      integer total, nbc

      nbc = 2*(mx + my)

      pcval        = 1
      ucval        = pcval        + nbc
      bcval        = ucval        + nbc
      trisys_du    = bcval        + 3*nbc
      trisys_d     = trisys_du    + nbc
      trisys_dl    = trisys_d     + nbc
      trisys_du2   = trisys_dl    + nbc
      qh           = trisys_du2   + nbc
      coeffs_alpha = qh           + nbc
      pval         = coeffs_alpha + 3
      dval         = pval         + nbc

c     # dval(2*(mx+my)+1)
      rhs          = dval         + nbc + 1
      rv           = rhs          + nbc
c     # rv(0:2*(mx+my))
      tv           = rv           + nbc + 1

c     # total  = 15*nbc + 6
      total        = tv           + nbc + 1

c     # Actual memory needed :
      total = total - 1

      if (mwork .lt. 0) then
         return
      endif
      if (mwork .lt. total) then
         write(6,*) 'Increase memory in work_tri to ',total
         stop
      endif

      end
