c     # -----------------------------------------------------
c     # parabolic_step_feuler
c     #
c     # Advance the parabolic step from time 't' to time 't + dt'
c     # using Forward Euler
c     #
c     #      nv(mq) = .true.  ==> diffuse species mq
c     #      nv(mq) = .false. ==> do not diffuse species mq
c     #                           mq = 1,..,meqn
c     #
c     # Normally, you have 'nv(mq) == plist(mq)'.  The only case
c     # in which this is not a good choice is the situation in which
c     # the sources terms containing the diffusion terms are
c     # completely decoupled.  In this case, it is much faster to
c     # solve each diffusion term separately.
c     #
c     # The variable 'ierror' will always return 0 (i.e. successful return)
c     #
c     # This routine is called either from the 'src2' routine in
c     # clawpack, or directly from 'parabolic_main' (if the clawpack
c     # driver is not being used.
c     # -----------------------------------------------------

      subroutine parabolic_step_feuler(maxmx,maxmy,meqn,mbc,
     &      mx,my,nv,q_claw,t,dt,ierror)
      implicit none

      external f_exp
      integer maxmx, maxmy, meqn, mbc,mx,my
      double precision q_claw(1-mbc:maxmx+mbc,1-mbc:maxmy+mbc,meqn)

      logical nv(meqn)
      double precision t, dt

      integer neqn, mc
      double precision ti, tip1

c     # Misc
      integer i,m, ntotal, ierror

      include "para.i"

      call check_method(method_feuler)

c     # We might have ntotal < meqn if not all terms are solved at once.
      call set_nv(meqn,nv,ntotal)

      if (ntotal .eq. 0) then
c        # We have no diffusion terms.
         ierror = 0
         return
      endif

      call ps_copy_from_claw(maxmx,maxmy,mbc,meqn,mx,my,
     &      q_claw,work(q_ptr))

      neqn = (mx+2)*(my+2)*ntotal

      ti = t
      tip1 = t + dt

c     # Do update based on explicit Euler steps using dt ~ dx^2, so that
c     # the scheme is second order.  We use the same workspace as that used
c     # of rkc, since rkc needs about 5 times as much space.
      call feuler(neqn,f_exp,work(q_ptr),ti,tip1,
     &      work(work_solver_ptr))

c     # What could go wrong with this scheme?
      ierror = 0

      call ps_copy_to_claw(maxmx,maxmy,mbc,meqn,mx,my,
     &      work(q_ptr),q_claw)

      end


      subroutine feuler(neqn,f_exp,q,ti,tip1,rhs)
      implicit none

      external f_exp
      integer neqn
      double precision ti,tip1
      double precision q(neqn)
      double precision rhs(neqn)

      double precision t_inner, dx,frac_dt,dt_exp, dt_stab
      double precision dt_inner
      integer m, dim

      include 'para.i'

c     # Get stable time step as use this as fixed step
      dx = dx_com

c     # Stable dt for constant coefficient.
      dim = 2
      dt_stab = dx*dx/(2*dim)

c     # Scaling factor on the stability limit
      frac_dt = 0.5d0
      dt_exp = frac_dt*dt_stab

      t_inner = ti

      do while (t_inner .lt. tip1)
         dt_inner = min(dt_exp, tip1 - t_inner)

c        # Use the same right hand side routine as for rkc
         call f_exp(neqn,t_inner,q,rhs)

         do m = 1,neqn
            q(m) = q(m) + dt_inner*rhs(m)
         enddo

         t_inner = t_inner + dt_inner
      enddo

      end

      subroutine feuler_stats(tip1,dt)
      implicit none

      double precision tip1, dt

      integer ts_counter
      integer get_time_step_counter

      integer lc_total, lc_inc
      integer get_laplacian_calls_elapsed
      integer get_laplacian_calls_total

      integer verbose
      integer get_verbosity

      lc_inc = get_laplacian_calls_elapsed()
      lc_total = get_laplacian_calls_total()
      verbose  = get_verbosity()
      ts_counter = get_time_step_counter()

      write(6,100) 'Step ', ts_counter
      write(6,101) 'Time ', tip1
      write(6,102) 'dt ',dt
      write(6,100) 'Laplacian calls (this step) ', lc_inc
      write(6,100) 'Laplacian calls (total) ', lc_total
      write(6,100) 'Number of internal time steps', lc_inc
      write(6,100)

  100 format(A35,I12)
  101 format(A35,F12.6)
  102 format(A35,E12.4)

      end
