!!       # -----------------------------------------------------
!!       # parabolic_step_irkc
!!       #
!!       # Advance the parabolic step from time 't' to time 't + dt'.
!!       #
!!       #      nv(mq) = .true.  ==> diffuse species mq
!!       #      nv(mq) = .false. ==> do not diffuse species mq
!!       # mq = 1,..,meqn
!!       #
!!       # Normally, you have 'nv(mq) == plist(mq)'.  The only case
!!       # in which this is not a good choice is the situation in which
!!       # the sources terms containing the diffusion terms are
!!       # completely decoupled.  In this case, it is much faster to
!!       # solve each diffusion term separately.
!!       #
!!       # After each call, Check ierror to make sure that the RKC
!!       # solver returned a valid solution.
!!       #
!!       #  ierror == 0   Normal exit
!!       #
!!       #  ierror  > 0   Problem with RKC solver.   Usually this
!!       #                means that your time step is too large
!!       #                for the requested accuracy.  See the
!!       #                description of 'idid' in the RKC documentation.
!!       #
!!       # This routine is called either from the 'src2' routine in
!!       # clawpack, or directly from 'parabolic_main' (if the clawpack
!!       # driver is not being used.
!!       # -----------------------------------------------------
SUBROUTINE parabolic_step_irkc(maxmx,maxmy,meqn,mbc,mx,my,nv,q_claw,t,dt,ierror)
  USE irkc_m

  IMPLICIT NONE

  EXTERNAL f_exp_irkc, f_imp_irkc, f_rkc
  INTEGER maxmx, maxmy, mbc,meqn,mx,my
  DOUBLE PRECISION q_claw(1-mbc:maxmx+mbc,1-mbc:maxmy+mbc,meqn)
  LOGICAL nv(meqn)
  DOUBLE PRECISION t, dt

  !! # parameters for the rkc solver
  DOUBLE PRECISION rtol, atol
  DOUBLE PRECISION ti, tip1
  DOUBLE PRECISION, ALLOCATABLE :: qwork(:)

  DOUBLE PRECISION dx, dy

  TYPE(IRKC_SOL) :: SOL

  !! # Misc
  INTEGER neqn, ntotal, ierror, ier, mpara

  INCLUDE "para.i"

  dx = dx_com
  dy = dy_com
  mpara = mpara_com

  call check_method(method_irkc)

  CALL set_nv(meqn,nv,ntotal)

  IF (ntotal .EQ. 0) THEN
     !! # We have no diffusion terms.
     ierror = 0
     RETURN
  ENDIF

  !!  # ntotal : Number of equations with parabolic terms. Note that
  !!  # ntotal <= meqn.
  neqn = (mx+2)*(my+2)*ntotal
  nactive_com = ntotal

  ALLOCATE(qwork(neqn), stat=ier)
   IF (ier .NE. 0) THEN
      WRITE(6,*) 'parabolic_step_irkc : problem with allocate'
      STOP
   ENDIF

  CALL ps_copy_from_claw_irkc(maxmx,maxmy,mbc,meqn,ntotal,mx,my,q_claw,qwork)

  rtol = 1e-4
  atol = 1e-4

  ti = t
  tip1 = t + dt

  SOL = IRKC_SET(ti,qwork,tip1,NPDES=ntotal,RE=rtol,AE=atol,ONE_STEP=.FALSE.,CONSTANT_J=.TRUE.,&
       STOP_ON_ERROR=.FALSE.)

  CALL IRKC(SOL,f_exp_irkc,f_imp_irkc)

  CALL set_irkc_stats(sol)

  ierror = sol%err_flag
  IF (ierror .NE. 0) THEN
     WRITE(6,*) 'parabolic_step_irkc : Problem with solver; ierror = ', ierror
     stop
  ENDIF

  CALL ps_copy_to_claw_irkc(maxmx,maxmy,mbc,meqn,ntotal,mx,my,sol%y,q_claw)

   DEALLOCATE(qwork,stat=ier)
   IF (ier .NE. 0) THEN
      WRITE(6,*) 'parabolic_step_irkc : problem with deallocate'
      STOP
   ENDIF

   DEALLOCATE(SOL%Y,SOL%YP,SOL%YLAST, SOL%YPLAST,STAT=IER)
   IF (ier .NE. 0) THEN
      WRITE(6,*) 'parabolic_step_irkc : problem with deallocate'
      STOP
   ENDIF


  RETURN
END SUBROUTINE parabolic_step_irkc



SUBROUTINE set_irkc_stats(sol)
  USE IRKC_m
  IMPLICIT NONE

  TYPE(IRKC_SOL) :: SOL
  INTEGER nfe,nfi,nsteps,naccpt,nrejct,nfesig,maxm,stats(6)

  !! # Common block for IRKC solver - this is set by solver.
  COMMON /irkcdid/   nfe,nfi,nsteps,naccpt,nrejct,nfesig,maxm

  nfe = sol%nfe
  nfi = sol%nfi
  nsteps = sol%nsteps
  naccpt = sol%naccpt
  nrejct = sol%nrejct
  nfesig = sol%nfesig
  maxm = sol%maxm

end




INTEGER FUNCTION get_stages_irkc()
  IMPLICIT NONE

  INTEGER nfe,nfi,nsteps,naccpt,nrejct,nfesig,maxm,stats(6)

  !!     # Common block for RKC solver - this is set by solver.
  COMMON /irkcdid/   nfe,nfi,nsteps,naccpt,nrejct,nfesig,maxm

  get_stages_irkc = maxm

END FUNCTION get_stages_irkc

INTEGER FUNCTION get_fevals_irkc()
  IMPLICIT NONE

  INTEGER nfe,nfi,nsteps,naccpt,nrejct,nfesig,maxm,stats(6)

  !!     # Common block for RKC solver - these are set by solver.
  COMMON /irkcdid/   nfe,nfi,nsteps,naccpt,nrejct,nfesig,maxm

  get_fevals_irkc = nfesig + nfe

END FUNCTION get_fevals_irkc


SUBROUTINE ps_copy_from_claw_irkc(maxmx, maxmy, mbc,meqn,nactive,mx,my,q_claw, q)
  IMPLICIT NONE

  INTEGER maxmx, maxmy, mbc,meqn,mx,my,nactive
  DOUBLE PRECISION q_claw(1-mbc:maxmx+mbc,1-mbc:maxmy+mbc,meqn)
  DOUBLE PRECISION q(nactive,0:mx+1,0:my+1)

  INTEGER i,j,m,mq
  LOGICAL active_parabolic_term
  INTEGER q_idx

  DO mq = 1,meqn
     IF (active_parabolic_term(mq)) THEN
        m = q_idx(mq)
        DO j = 0,my+1
           DO i = 0,mx+1
              q(m,i,j) = q_claw(i,j,mq)
           ENDDO
        ENDDO
     END IF
  ENDDO

END SUBROUTINE ps_copy_from_claw_irkc

SUBROUTINE ps_copy_to_claw_irkc(maxmx, maxmy, mbc,meqn,nactive,mx,my,q,q_claw)
  IMPLICIT NONE

  INTEGER maxmx, maxmy, mbc,meqn,mx,my, nv,nactive
  DOUBLE PRECISION q(nactive,0:mx+1,0:my+1)
  DOUBLE PRECISION q_claw(1-mbc:maxmx+mbc,1-mbc:maxmy+mbc,meqn)

  INTEGER i,j, m,mq
  LOGICAL active_parabolic_term
  INTEGER q_idx

  DO mq = 1,meqn
     IF (active_parabolic_term(mq)) THEN
        m = q_idx(mq)
        DO j = 0,my+1
           DO i = 0,mx+1
              q_claw(i,j,mq) = q(m,i,j)
           END DO
        ENDDO
     END IF
  ENDDO

END SUBROUTINE ps_copy_to_claw_irkc



SUBROUTINE irkc_stats(tip1,dt)
  IMPLICIT NONE

  DOUBLE PRECISION tip1,dt

  INTEGER ts_counter
  INTEGER get_time_step_counter

  INTEGER lc_total, lc_inc
  INTEGER get_laplacian_calls_total
  INTEGER get_laplacian_calls_elapsed

  INTEGER mx, my, mpara, neqn
  INTEGER nfe,nfi,nsteps,naccpt,nrejct,nfesig,maxm,stats(6)

  !!     # COMMON BLOCK for RKC solver - this is set by solver.
  COMMON /irkcdid/   nfe,nfi,nsteps,naccpt,nrejct,nfesig,maxm

  INCLUDE 'para.i'

  mx = mx_com
  my = my_com
  mpara = mpara_com
  neqn = (mx+2)*(my+2)*mpara


  ts_counter = get_time_step_counter()
  lc_total = get_laplacian_calls_total()
  lc_inc = get_laplacian_calls_elapsed()

  CALL write_line(6,47)
  WRITE(6,100) 'Step ', ts_counter
  WRITE(6,101) 'Time ', tip1
  WRITE(6,102) 'dt ',dt
  WRITE(6,100) 'Number of Lap. calls (inc)', lc_inc
  WRITE(6,100) 'Number of Lap. calls (total)', lc_total

  WRITE(6,100) 'Number of explicit f evaluations', nfe
  WRITE(6,100) 'Number of implicit f evals/grd pt', nfi/neqn
  WRITE(6,100) 'Number of steps', nsteps
  WRITE(6,100) 'Number of accepted steps', naccpt
  WRITE(6,100) 'Number of rejected steps', nrejct
  WRITE(6,100) 'Number of evals. for Jacobian', nfesig
  WRITE(6,100) 'Maximum number of stages used', maxm
  WRITE(6,100) 'Number of fevals (curr. step)', nfesig + nfe
  WRITE(6,100) 'Total number of fevals', lc_total

  CALL write_line(6,47)
  WRITE(6,*) ' '


100 FORMAT(A35,I12)
101 FORMAT(A35,F12.6)
102 FORMAT(A35,E12.4)

END SUBROUTINE irkc_stats
