c     # -----------------------------------------------------
c     # parabolic_step_rock4
c     #
c     # Advance the parabolic step from time 't' to time 't + dt'.
c     #
c     #      nv(mq) = .true.  ==> diffuse species mq
c     #      nv(mq) = .false. ==> do not diffuse species mq
c     # mq = 1,..,meqn
c     #
c     # Normally, you have 'nv(mq) == plist(mq)'.  The only case
c     # in which this is not a good choice is the situation in which
c     # the sources terms containing the diffusion terms are
c     # completely decoupled.  In this case, it is much faster to
c     # solve each diffusion term separately.
c     #
c     # After each call, Check ierror to make sure that the RKC
c     # solver returned a valid solution.
c     #
c     #  ierror == 0   Normal exit
c     #
c     #  ierror  > 0   Problem with RKC solver.   Usually this
c     #                means that your time step is too large
c     #                for the requested accuracy.  See the
c     #                description of 'idid' in the RKC documentation.
c     #
c     # This routine is called either from the 'src2' routine in
c     # clawpack, or directly from 'parabolic_main' (if the clawpack
c     # driver is not being used.
c     # -----------------------------------------------------
      subroutine parabolic_step_rock4(maxmx,maxmy,meqn,mbc,
     &      mx,my,nv,q_claw,t,dt,ierror)
      implicit none

      external f_exp
      integer maxmx, maxmy, mbc,meqn,mx,my
      double precision q_claw(1-mbc:maxmx+mbc,1-mbc:maxmy+mbc,meqn)
      logical nv(meqn)
      double precision t, dt

c     # parameters for the rkc solver
      integer neqn, iwork(12), idid, iwork_com(12)
      double precision rtol, atol
      double precision ti, tip1, dt_init

c     # Misc
      integer i,m, ntotal, ierror

      common /comrock/ iwork_com

      include "para.i"

      call check_method(method_rock4)

      call set_nv(meqn,nv,ntotal)

      if (ntotal .eq. 0) then
c        # We have no diffusion terms.
         ierror = 0
         return
      endif

      call ps_copy_from_claw(maxmx,maxmy,mbc,meqn,mx,my,
     &      q_claw,work(q_ptr))

c     # ntotal : Number of equations with parabolic terms. Note that
c     # ntotal <= meqn.
      neqn = (mx+2)*(my+2)*ntotal

      rtol = 1d-8
      atol = 1d-8

      iwork(1) = 0
      iwork(2) = 0
      iwork(3) = 0
      iwork(4) = 0

c     # Take one step of the ROCK4 solver.  We rely on the user to
c     # specify an appropriate time step.   If they see that
c     # ierror /= 0, then they should adjust the time step
c     # accordingly.
      ti = t
      tip1 = t + dt

      dt_init = 0.01*dt

      idid = 0
      ierror = 0
      call rock4(neqn,ti,tip1,dt_init,work(q_ptr),f_exp,
     &      atol,rtol,work(work_solver_ptr),iwork,idid)

c     # 'ierror' value should be checked by the user on return
      if (idid .le. 1) then
         ierror = 0
         do m = 1,12
            iwork_com(m) = iwork(m)
         enddo
      else
         write(6,*) 'parabolic_step_rock4 : Problem with solution; ',
     &         'idid = ', idid
         ierror = 1
      endif

      call ps_copy_to_claw(maxmx,maxmy,mbc,meqn,mx,my,
     &      work(q_ptr),q_claw)

      return
      end


      subroutine rock4_stats(tip1,dt)
      implicit none

      double precision tip1,dt

      integer ts_counter
      integer get_time_step_counter

      integer lc_total, lc_inc
      integer get_laplacian_calls_total
      integer get_laplacian_calls_elapsed

      integer iwork(12)
      common /comrock/ iwork

      ts_counter = get_time_step_counter()
      lc_total = get_laplacian_calls_total()
      lc_inc = get_laplacian_calls_elapsed()

c      nf_total = lc_total + iwork(9) + iwork(5)

      call write_line(6,47)
      write(6,100) 'Step ', ts_counter
      write(6,101) 'Time ', tip1
      write(6,102) 'dt ',dt
      write(6,100) 'Number of Lap. calls (inc)', lc_inc
      write(6,100) 'Number of Lap. calls (total)', lc_total

      write(6,100) 'Number of f evaluations ', iwork(5)
      write(6,100) 'Number of steps ', iwork(6)
      write(6,100) 'Number of accepted steps ', iwork(7)
      write(6,100) 'Number of rejected steps ', iwork(8)
      write(6,100) 'Number of evals. for Jacobian ', iwork(9)
      write(6,100) 'Maximum number of stages used ', iwork(10)
      write(6,100) 'Number of fevals (curr. step) ', iwork(9) + iwork(5)
      write(6,100) 'Total number of fevals ', lc_total

      call write_line(6,47)
      write(6,*) ' '


  100 format(A35,I12)
  101 format(A35,F12.6)
  102 format(A35,E12.4)

      end

      subroutine rock4_stats_short(tip1,dt,step_counter)
      implicit none

      double precision tip1,dt
      integer step_counter

      integer get_stages_rock4, get_fevals_rock4
      integer istages, ifevals

      istages = get_stages_rock4()
      ifevals = get_fevals_rock4()
      write(6,500) step_counter, tip1, dt, istages, ifevals
  500 format('Step ', i4,
     &      ' at time t = ',F12.4,
     &      '   (dt = ',E10.4,
     &      ';  max stages = ',I3,
     &      ';  fevals = ',I4,')')

      end


      integer function get_stages_rock4()
      implicit none

      integer iwork(12)
      common /comrock/ iwork

      get_stages_rock4 = iwork(10)

      end

      integer function get_fevals_rock4()
      implicit none

      integer iwork(12)
      common /comrock/ iwork

      get_fevals_rock4 = iwork(9) + iwork(5)

      end


      double precision function rho(neqn,t,y)
      implicit none
      integer          neqn
      double precision t,y(neqn)

      rho = 0

      return
      end
